# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import tarfile
import warnings
from collections.abc import Callable, Iterator
from io import BytesIO
from pathlib import Path
from string import hexdigits
from time import time
from typing import TYPE_CHECKING, Optional, overload

# Import from pygit2
from ._pygit2 import (
    GIT_OID_HEXSZ,
    GIT_OID_MINPREFIXLEN,
    Blob,
    Commit,
    Diff,
    InvalidSpecError,
    Object,
    Oid,
    Patch,
    Reference,
    Signature,
    Tree,
    init_file_backend,
)
from ._pygit2 import Repository as _Repository
from .blame import Blame
from .branches import Branches
from .callbacks import (
    StashApplyCallbacks,
    git_checkout_options,
    git_stash_apply_options,
)
from .config import Config
from .enums import (
    AttrCheck,
    BlameFlag,
    CheckoutStrategy,
    DescribeStrategy,
    DiffOption,
    FileMode,
    FilterMode,
    MergeFavor,
    MergeFileFlag,
    MergeFlag,
    ObjectType,
    RepositoryOpenFlag,
    RepositoryState,
)
from .errors import check_error
from .ffi import C, ffi
from .filter import FilterList
from .index import Index, IndexEntry, MergeFileResult
from .packbuilder import PackBuilder
from .references import References
from .remotes import RemoteCollection
from .submodules import SubmoduleCollection
from .transaction import ReferenceTransaction
from .utils import StrArray, to_bytes

if TYPE_CHECKING:
    from pygit2._libgit2.ffi import (
        ArrayC,
        GitMergeOptionsC,
        GitRepositoryC,
        _Pointer,
        char,
    )
    from pygit2._pygit2 import Odb, Refdb, RefdbBackend


class BaseRepository(_Repository):
    _pointer: '_Pointer[GitRepositoryC]'
    _repo: 'GitRepositoryC'
    backend: 'RefdbBackend'
    default_signature: Signature
    head: Reference
    head_is_detached: bool
    head_is_unborn: bool
    is_bare: bool
    is_empty: bool
    is_shallow: bool
    odb: 'Odb'
    path: str
    refdb: 'Refdb'
    workdir: str
    references: References
    remotes: RemoteCollection
    branches: Branches
    submodules: SubmoduleCollection

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._common_init()

    def _common_init(self) -> None:
        self.branches = Branches(self)
        self.references = References(self)
        self.remotes = RemoteCollection(self)
        self.submodules = SubmoduleCollection(self)
        self._active_transaction = None

        # Get the pointer as the contents of a buffer and store it for
        # later access
        repo_cptr = ffi.new('git_repository **')
        ffi.buffer(repo_cptr)[:] = self._pointer[:]
        self._repo = repo_cptr[0]

    # Backwards compatible ODB access
    def read(self, oid: Oid | str) -> tuple[int, bytes]:
        """read(oid) -> type, data, size

        Read raw object data from the repository.
        """
        return self.odb.read(oid)

    def write(self, type: int, data: bytes | str) -> Oid:
        """write(type, data) -> Oid

        Write raw object data into the repository. First arg is the object
        type, the second one a buffer with data. Return the Oid of the created
        object."""
        return self.odb.write(type, data)

    def pack(
        self,
        path: str | Path | None = None,
        pack_delegate: Callable[[PackBuilder], None] | None = None,
        n_threads: int | None = None,
    ) -> int:
        """Pack the objects in the odb chosen by the pack_delegate function
        and write `.pack` and `.idx` files for them.

        Returns: the number of objects written to the pack

        Parameters:

        path
            The path to which the `.pack` and `.idx` files should be written. `None` will write to the default location.

        pack_delegate
            The method which will provide add the objects to the pack builder. Defaults to all objects.

        n_threads
            The number of threads the `PackBuilder` will spawn. If set to 0, libgit2 will autodetect the number of CPUs.
        """

        def pack_all_objects(pack_builder):
            for obj in self.odb:
                pack_builder.add(obj)

        pack_delegate = pack_delegate or pack_all_objects

        builder = PackBuilder(self)
        if n_threads is not None:
            builder.set_threads(n_threads)
        pack_delegate(builder)
        builder.write(path=path)

        return builder.written_objects_count

    def hashfile(
        self,
        path: str,
        object_type: ObjectType = ObjectType.BLOB,
        as_path: str | None = None,
    ) -> Oid:
        """Calculate the hash of a file using repository filtering rules.

        If you simply want to calculate the hash of a file on disk with no filters,
        you can just use `pygit2.hashfile()`. However, if you want to hash a file
        in the repository and you want to apply filtering rules (e.g. crlf filters)
        before generating the SHA, then use this function.

        Note: if the repository has `core.safecrlf` set to fail and the filtering
        triggers that failure, then this function will raise an error and not
        calculate the hash of the file.

        Returns: Output value of calculated SHA (Oid)

        Parameters:

        path
            Path to file on disk whose contents should be hashed. This may be
            an absolute path or a relative path, in which case it will be treated
            as a path within the working directory.

        object_type
            The object type to hash (e.g. enums.ObjectType.BLOB)

        as_path
            The path to use to look up filtering rules. If this is an empty string
            then no filters will be applied when calculating the hash.
            If this is `None` and the `path` parameter is a file within the
            repository's working directory, then the `path` will be used.
        """
        c_path = to_bytes(path)

        c_as_path: ffi.NULL_TYPE | bytes
        if as_path is None:
            c_as_path = ffi.NULL
        else:
            c_as_path = to_bytes(as_path)

        c_oid = ffi.new('git_oid *')

        err = C.git_repository_hashfile(
            c_oid, self._repo, c_path, int(object_type), c_as_path
        )
        check_error(err)

        oid = Oid(raw=bytes(ffi.buffer(c_oid.id)[:]))
        return oid

    def load_filter_list(
        self, path: str, mode: FilterMode = FilterMode.TO_ODB
    ) -> FilterList | None:
        """
        Load the filter list for a given path.
        May return None if there are no filters to apply to this path.

        Parameters:

        path
            Relative path of the file to be filtered

        mode
            Filtering direction: ODB to worktree (SMUDGE), or worktree to ODB
            (CLEAN).
        """
        c_filters = ffi.new('git_filter_list **')
        c_path = to_bytes(path)
        c_mode = int(mode)

        err = C.git_filter_list_load(c_filters, self._repo, ffi.NULL, c_path, c_mode, 0)
        check_error(err)
        fl = FilterList._from_c(c_filters[0])
        return fl

    def __iter__(self) -> Iterator[Oid]:
        return iter(self.odb)

    #
    # Mapping interface
    #
    def get(self, key: Oid | str, default: Optional[Commit] = None) -> None | Object:
        value = self.git_object_lookup_prefix(key)
        return value if (value is not None) else default

    def __getitem__(self, key: str | Oid) -> Object:
        value = self.git_object_lookup_prefix(key)
        if value is None:
            raise KeyError(key)
        return value

    def __contains__(self, key: str | Oid) -> bool:
        return self.git_object_lookup_prefix(key) is not None

    def __repr__(self) -> str:
        return f'pygit2.Repository({repr(self.path)})'

    #
    # Configuration
    #
    @property
    def config(self) -> Config:
        """The configuration file for this repository.

        If a the configuration hasn't been set yet, the default config for
        repository will be returned, including global and system configurations
        (if they are available).
        """
        cconfig = ffi.new('git_config **')
        err = C.git_repository_config(cconfig, self._repo)
        check_error(err)

        return Config.from_c(self, cconfig[0])

    @property
    def config_snapshot(self):
        """A snapshot for this repositiory's configuration

        This allows reads over multiple values to use the same version
        of the configuration files.
        """
        cconfig = ffi.new('git_config **')
        err = C.git_repository_config_snapshot(cconfig, self._repo)
        check_error(err)

        return Config.from_c(self, cconfig[0])

    #
    # References
    #
    def create_reference(
        self,
        name: str,
        target: Oid | str,
        force: bool = False,
        message: str | None = None,
    ) -> 'Reference':
        """Create a new reference "name" which points to an object or to
        another reference.

        Based on the type and value of the target parameter, this method tries
        to guess whether it is a direct or a symbolic reference.

        Keyword arguments:

        force: bool
            If True references will be overridden, otherwise (the default) an
            exception is raised.

        message: str
            Optional message to use for the reflog.

        Examples::

            repo.create_reference('refs/heads/foo', repo.head.target)
            repo.create_reference('refs/tags/foo', 'refs/heads/master')
            repo.create_reference('refs/tags/foo', 'bbb78a9cec580')
        """
        direct = isinstance(target, Oid) or (
            all(c in hexdigits for c in target)
            and GIT_OID_MINPREFIXLEN <= len(target) <= GIT_OID_HEXSZ
        )

        # duplicate isinstance call for mypy
        if direct or isinstance(target, Oid):
            return self.create_reference_direct(name, target, force, message=message)

        return self.create_reference_symbolic(name, target, force, message=message)

    def listall_references(self) -> list[str]:
        """Return a list with all the references in the repository."""
        return list(x.name for x in self.references.iterator())

    def listall_reference_objects(self) -> list[Reference]:
        """Return a list with all the reference objects in the repository."""
        return list(x for x in self.references.iterator())

    def resolve_refish(self, refish: str) -> tuple[Commit, Reference]:
        """Convert a reference-like short name "ref-ish" to a valid
        (commit, reference) pair.

        If ref-ish points to a commit, the reference element of the result
        will be None.

        Examples::

            repo.resolve_refish('mybranch')
            repo.resolve_refish('sometag')
            repo.resolve_refish('origin/master')
            repo.resolve_refish('bbb78a9')
        """
        try:
            reference = self.lookup_reference_dwim(refish)
        except (KeyError, InvalidSpecError):
            reference = None
            commit = self.revparse_single(refish)
        else:
            commit = reference.peel(Commit)  # type: ignore

        return (commit, reference)  # type: ignore

    def transaction(self) -> ReferenceTransaction:
        """Create a new reference transaction.

        Returns a context manager that commits all reference updates atomically
        when the context exits successfully, or performs no updates if an exception
        is raised.

        Example::

            with repo.transaction() as txn:
                txn.lock_ref('refs/heads/master')
                txn.set_target('refs/heads/master', new_oid, message='Update')
        """
        txn = ReferenceTransaction(self)
        return txn

    #
    # Checkout
    #

    def checkout_head(self, **kwargs):
        """Checkout HEAD

        For arguments, see Repository.checkout().
        """
        with git_checkout_options(**kwargs) as payload:
            err = C.git_checkout_head(self._repo, payload.checkout_options)
            payload.check_error(err)

    def checkout_index(self, index=None, **kwargs):
        """Checkout the given index or the repository's index

        For arguments, see Repository.checkout().
        """
        with git_checkout_options(**kwargs) as payload:
            err = C.git_checkout_index(
                self._repo,
                index._index if index else ffi.NULL,
                payload.checkout_options,
            )
            payload.check_error(err)

    def checkout_tree(self, treeish, **kwargs):
        """Checkout the given treeish

        For arguments, see Repository.checkout().
        """
        with git_checkout_options(**kwargs) as payload:
            cptr = ffi.new('git_object **')
            ffi.buffer(cptr)[:] = treeish._pointer[:]
            err = C.git_checkout_tree(self._repo, cptr[0], payload.checkout_options)
            payload.check_error(err)

    def checkout(
        self,
        refname: str | None | Reference = None,
        **kwargs,
    ) -> None:
        """
        Checkout the given reference using the given strategy, and update the
        HEAD.
        The reference may be a reference name or a Reference object.
        The default strategy is SAFE | RECREATE_MISSING.

        If no reference is given, checkout from the index.

        Parameters:

        refname : str or Reference
            The reference to checkout. After checkout, the current branch will
            be switched to this one.

        strategy : CheckoutStrategy
            A ``CheckoutStrategy`` value. The default is ``SAFE | RECREATE_MISSING``.

        directory : str
            Alternative checkout path to workdir.

        paths : list[str]
            A list of files to checkout from the given reference.
            If paths is provided, HEAD will not be set to the reference.

        callbacks : CheckoutCallbacks
            Optional. Supply a `callbacks` object to get information about
            conflicted files, updated files, etc. as the checkout is being
            performed. The callbacks can also abort the checkout prematurely.

            The callbacks should be an object which inherits from
            `pyclass:CheckoutCallbacks`. It should implement the callbacks
            as overridden methods.

        Examples:

        * To checkout from the HEAD, just pass 'HEAD'::

            >>> checkout('HEAD')

          This is identical to calling checkout_head().
        """

        # Case 1: Checkout index
        if refname is None:
            return self.checkout_index(**kwargs)

        # Case 2: Checkout head
        if refname == 'HEAD':
            return self.checkout_head(**kwargs)

        # Case 3: Reference
        if isinstance(refname, Reference):
            reference = refname
            refname = refname.name
        else:
            reference = self.lookup_reference(refname)

        oid = reference.resolve().target
        treeish = self[oid]
        self.checkout_tree(treeish, **kwargs)

        if 'paths' not in kwargs:
            self.set_head(refname)

    #
    # Setting HEAD
    #
    def set_head(self, target: Oid | str) -> None:
        """
        Set HEAD to point to the given target.

        Parameters:

        target
            The new target for HEAD. Can be a string or Oid (to detach).
        """

        if isinstance(target, Oid):
            oid = ffi.new('git_oid *')
            ffi.buffer(oid)[:] = target.raw[:]
            err = C.git_repository_set_head_detached(self._repo, oid)
            check_error(err)
            return

        # if it's a string, then it's a reference name
        err = C.git_repository_set_head(self._repo, to_bytes(target))
        check_error(err)

    #
    # Diff
    #
    def __whatever_to_tree_or_blob(self, obj):
        if obj is None:
            return None

        # If it's a string, then it has to be valid revspec
        if isinstance(obj, str) or isinstance(obj, bytes):
            obj = self.revparse_single(obj)
        elif isinstance(obj, Oid):
            obj = self[obj]

        # First we try to get to a blob
        try:
            obj = obj.peel(Blob)
        except Exception:
            # And if that failed, try to get a tree, raising a type
            # error if that still doesn't work
            try:
                obj = obj.peel(Tree)
            except Exception:
                raise TypeError(f'unexpected "{type(obj)}"')

        return obj

    @overload
    def diff(
        self,
        a: None | str | bytes | Commit | Oid | Reference = None,
        b: None | str | bytes | Commit | Oid | Reference = None,
        cached: bool = False,
        flags: DiffOption = DiffOption.NORMAL,
        context_lines: int = 3,
        interhunk_lines: int = 0,
    ) -> Diff: ...
    @overload
    def diff(
        self,
        a: Blob | None = None,
        b: Blob | None = None,
        cached: bool = False,
        flags: DiffOption = DiffOption.NORMAL,
        context_lines: int = 3,
        interhunk_lines: int = 0,
    ) -> Patch: ...
    def diff(
        self,
        a: None | Blob | str | bytes | Commit | Oid | Reference = None,
        b: None | Blob | str | bytes | Commit | Oid | Reference = None,
        cached: bool = False,
        flags: DiffOption = DiffOption.NORMAL,
        context_lines: int = 3,
        interhunk_lines: int = 0,
    ) -> Diff | Patch:
        """
        Show changes between the working tree and the index or a tree,
        changes between the index and a tree, changes between two trees, or
        changes between two blobs.

        Keyword arguments:

        a
            None, a str (that refers to an Object, see revparse_single()) or a
            Reference object.
            If None, b must be None, too. In this case the working directory is
            compared with the index. Otherwise the referred object is compared to
            'b'.

        b
            None, a str (that refers to an Object, see revparse_single()) or a
            Reference object.
            If None, the working directory is compared to 'a'. (except
            'cached' is True, in which case the index is compared to 'a').
            Otherwise the referred object is compared to 'a'

        cached
            If 'b' is None, by default the working directory is compared to 'a'.
            If 'cached' is set to True, the index/staging area is used for comparing.

        flag
            A combination of enums.DiffOption constants.

        context_lines
            The number of unchanged lines that define the boundary of a hunk
            (and to display before and after)

        interhunk_lines
            The maximum number of unchanged lines between hunk boundaries
            before the hunks will be merged into a one

        Examples::

          # Changes in the working tree not yet staged for the next commit
          >>> diff()

          # Changes between the index and your last commit
          >>> diff(cached=True)

          # Changes in the working tree since your last commit
          >>> diff('HEAD')

          # Changes between commits
          >>> t0 = revparse_single('HEAD')
          >>> t1 = revparse_single('HEAD^')
          >>> diff(t0, t1)
          >>> diff('HEAD', 'HEAD^') # equivalent

        If you want to diff a tree against an empty tree, use the low level
        API (Tree.diff_to_tree()) directly.
        """

        a = self.__whatever_to_tree_or_blob(a)
        b = self.__whatever_to_tree_or_blob(b)

        options = {
            'flags': int(flags),
            'context_lines': context_lines,
            'interhunk_lines': interhunk_lines,
        }

        # Case 1: Diff tree to tree
        if isinstance(a, Tree) and isinstance(b, Tree):
            return a.diff_to_tree(b, **options)  # type: ignore[arg-type]

        # Case 2: Index to workdir
        elif a is None and b is None:
            return self.index.diff_to_workdir(**options)

        # Case 3: Diff tree to index or workdir
        elif isinstance(a, Tree) and b is None:
            if cached:
                return a.diff_to_index(self.index, **options)  # type: ignore[arg-type]
            else:
                return a.diff_to_workdir(**options)  # type: ignore[arg-type]

        # Case 4: Diff blob to blob
        if isinstance(a, Blob) and isinstance(b, Blob):
            return a.diff(b, **options)  # type: ignore[arg-type]

        raise ValueError('Only blobs and treeish can be diffed')

    def state(self) -> RepositoryState:
        """Determines the state of a git repository - ie, whether an operation
        (merge, cherry-pick, etc) is in progress.

        Returns a RepositoryState constant.
        """
        cstate: int = C.git_repository_state(self._repo)
        try:
            return RepositoryState(cstate)
        except ValueError:
            # Some value not in the IntEnum - newer libgit2 version?
            return cstate  # type: ignore[return-value]

    def state_cleanup(self) -> None:
        """Remove all the metadata associated with an ongoing command like
        merge, revert, cherry-pick, etc. For example: MERGE_HEAD, MERGE_MSG,
        etc.
        """
        C.git_repository_state_cleanup(self._repo)

    #
    # blame
    #
    def blame(
        self,
        path: str,
        flags: BlameFlag = BlameFlag.NORMAL,
        min_match_characters: int | None = None,
        newest_commit: Oid | str | None = None,
        oldest_commit: Oid | str | None = None,
        min_line: int | None = None,
        max_line: int | None = None,
    ) -> Blame:
        """
        Return a Blame object for a single file.

        Parameters:

        path
            Path to the file to blame.

        flags
            An enums.BlameFlag constant.

        min_match_characters
            The number of alphanum chars that must be detected as moving/copying
            within a file for it to associate those lines with the parent commit.

        newest_commit
            The id of the newest commit to consider.

        oldest_commit
            The id of the oldest commit to consider.

        min_line
            The first line in the file to blame.

        max_line
            The last line in the file to blame.

        Examples::

            repo.blame('foo.c', flags=enums.BlameFlag.IGNORE_WHITESPACE)
        """

        options = ffi.new('git_blame_options *')

        C.git_blame_options_init(options, C.GIT_BLAME_OPTIONS_VERSION)
        if flags:
            options.flags = int(flags)
        if min_match_characters:
            options.min_match_characters = min_match_characters
        if newest_commit:
            if not isinstance(newest_commit, Oid):
                newest_commit = Oid(hex=newest_commit)
            ffi.buffer(ffi.addressof(options, 'newest_commit'))[:] = newest_commit.raw
        if oldest_commit:
            if not isinstance(oldest_commit, Oid):
                oldest_commit = Oid(hex=oldest_commit)
            ffi.buffer(ffi.addressof(options, 'oldest_commit'))[:] = oldest_commit.raw
        if min_line:
            options.min_line = min_line
        if max_line:
            options.max_line = max_line

        cblame = ffi.new('git_blame **')
        err = C.git_blame_file(cblame, self._repo, to_bytes(path), options)
        check_error(err)

        return Blame._from_c(self, cblame[0])

    #
    # Index
    #
    @property
    def index(self):
        """Index representing the repository's index file."""
        cindex = ffi.new('git_index **')
        err = C.git_repository_index(cindex, self._repo)
        check_error(err, io=True)

        return Index.from_c(self, cindex)

    #
    # Merging
    #
    @staticmethod
    def _merge_options(
        favor: int | MergeFavor, flags: int | MergeFlag, file_flags: int | MergeFileFlag
    ) -> 'GitMergeOptionsC':
        """Return a 'git_merge_opts *'"""

        # Check arguments type
        if not isinstance(favor, (int, MergeFavor)):
            raise TypeError('favor argument must be MergeFavor')

        if not isinstance(flags, (int, MergeFlag)):
            raise TypeError('flags argument must be MergeFlag')

        if not isinstance(file_flags, (int, MergeFileFlag)):
            raise TypeError('file_flags argument must be MergeFileFlag')

        opts = ffi.new('git_merge_options *')
        err = C.git_merge_options_init(opts, C.GIT_MERGE_OPTIONS_VERSION)
        check_error(err)

        opts.file_favor = int(favor)
        opts.flags = int(flags)
        opts.file_flags = int(file_flags)

        return opts

    def merge_file_from_index(
        self,
        ancestor: 'IndexEntry | None',
        ours: 'IndexEntry | None',
        theirs: 'IndexEntry | None',
        use_deprecated: bool = True,
    ) -> 'str | MergeFileResult | None':
        """Merge files from index.

        Returns: A string with the content of the file containing
        possible conflicts if use_deprecated==True.
        If use_deprecated==False then it returns an instance of MergeFileResult.

        ancestor
            The index entry which will be used as a common
            ancestor.
        ours
            The index entry to take as "ours" or base.
        theirs
            The index entry which will be merged into "ours"
        use_deprecated
            This controls what will be returned. If use_deprecated==True (default),
            a string with the contents of the file will be returned.
            An instance of MergeFileResult will be returned otherwise.
        """
        cmergeresult = ffi.new('git_merge_file_result *')

        cancestor, ancestor_str_ref = (
            ancestor._to_c() if ancestor is not None else (ffi.NULL, ffi.NULL)
        )
        cours, ours_str_ref = ours._to_c() if ours is not None else (ffi.NULL, ffi.NULL)
        ctheirs, theirs_str_ref = (
            theirs._to_c() if theirs is not None else (ffi.NULL, ffi.NULL)
        )

        err = C.git_merge_file_from_index(
            cmergeresult, self._repo, cancestor, cours, ctheirs, ffi.NULL
        )
        check_error(err)

        mergeFileResult = MergeFileResult._from_c(cmergeresult)
        C.git_merge_file_result_free(cmergeresult)

        if use_deprecated:
            warnings.warn(
                'Getting an str from Repository.merge_file_from_index is deprecated. '
                'The method will later return an instance of MergeFileResult by default, instead. '
                'Check parameter use_deprecated.',
                DeprecationWarning,
            )
            return mergeFileResult.contents if mergeFileResult else ''

        return mergeFileResult

    def merge_commits(
        self,
        ours: str | Oid | Commit,
        theirs: str | Oid | Commit,
        favor: MergeFavor = MergeFavor.NORMAL,
        flags: MergeFlag = MergeFlag.FIND_RENAMES,
        file_flags: MergeFileFlag = MergeFileFlag.DEFAULT,
    ) -> 'Index':
        """
        Merge two arbitrary commits.

        Returns: an index with the result of the merge.

        Parameters:

        ours
            The commit to take as "ours" or base.

        theirs
            The commit which will be merged into "ours"

        favor
            An enums.MergeFavor constant specifying how to deal with file-level conflicts.
            For all but NORMAL, the index will not record a conflict.

        flags
            A combination of enums.MergeFlag constants.

        file_flags
            A combination of enums.MergeFileFlag constants.

        Both "ours" and "theirs" can be any object which peels to a commit or
        the id (string or Oid) of an object which peels to a commit.
        """
        ours_ptr = ffi.new('git_commit **')
        theirs_ptr = ffi.new('git_commit **')
        cindex = ffi.new('git_index **')

        if isinstance(ours, (str, Oid)):
            ours_object = self[ours]
            if not isinstance(ours_object, Commit):
                raise TypeError(f'expected Commit, got {type(ours_object)}')
            ours = ours_object
        if isinstance(theirs, (str, Oid)):
            theirs_object = self[theirs]
            if not isinstance(theirs_object, Commit):
                raise TypeError(f'expected Commit, got {type(theirs_object)}')
            theirs = theirs_object

        ours = ours.peel(Commit)
        theirs = theirs.peel(Commit)

        opts = self._merge_options(favor, flags, file_flags)

        ffi.buffer(ours_ptr)[:] = ours._pointer[:]
        ffi.buffer(theirs_ptr)[:] = theirs._pointer[:]

        err = C.git_merge_commits(cindex, self._repo, ours_ptr[0], theirs_ptr[0], opts)
        check_error(err)

        return Index.from_c(self, cindex)

    def merge_trees(
        self,
        ancestor: str | Oid | Tree,
        ours: str | Oid | Tree,
        theirs: str | Oid | Tree,
        favor: MergeFavor = MergeFavor.NORMAL,
        flags: MergeFlag = MergeFlag.FIND_RENAMES,
        file_flags: MergeFileFlag = MergeFileFlag.DEFAULT,
    ) -> 'Index':
        """
        Merge two trees.

        Returns: an Index that reflects the result of the merge.

        Parameters:

        ancestor
            The tree which is the common ancestor between 'ours' and 'theirs'.

        ours
            The commit to take as "ours" or base.

        theirs
            The commit which will be merged into "ours".

        favor
            An enums.MergeFavor constant specifying how to deal with file-level conflicts.
            For all but NORMAL, the index will not record a conflict.

        flags
            A combination of enums.MergeFlag constants.

        file_flags
            A combination of enums.MergeFileFlag constants.
        """
        ancestor_ptr = ffi.new('git_tree **')
        ours_ptr = ffi.new('git_tree **')
        theirs_ptr = ffi.new('git_tree **')
        cindex = ffi.new('git_index **')

        ancestor = self.__ensure_tree(ancestor)
        ours = self.__ensure_tree(ours)
        theirs = self.__ensure_tree(theirs)

        opts = self._merge_options(favor, flags, file_flags)

        ffi.buffer(ancestor_ptr)[:] = ancestor._pointer[:]
        ffi.buffer(ours_ptr)[:] = ours._pointer[:]
        ffi.buffer(theirs_ptr)[:] = theirs._pointer[:]

        err = C.git_merge_trees(
            cindex, self._repo, ancestor_ptr[0], ours_ptr[0], theirs_ptr[0], opts
        )
        check_error(err)

        return Index.from_c(self, cindex)

    def merge(
        self,
        source: Reference | Commit | Oid | str,
        favor: MergeFavor = MergeFavor.NORMAL,
        flags: MergeFlag = MergeFlag.FIND_RENAMES,
        file_flags: MergeFileFlag = MergeFileFlag.DEFAULT,
    ) -> None:
        """
        Merges the given Reference or Commit into HEAD.

        Merges the given commit into HEAD, writing the results into the working directory.
        Any changes are staged for commit and any conflicts are written to the index.
        Callers should inspect the repository's index after this completes,
        resolve any conflicts and prepare a commit.

        Parameters:

        source
            The Reference, Commit, or commit Oid to merge into HEAD.
            It is preferable to pass in a Reference, because this enriches the
            merge with additional information (for example, Repository.message will
            specify the name of the branch being merged).
            Previous versions of pygit2 allowed passing in a partial commit
            hash as a string; this is deprecated.

        favor
            An enums.MergeFavor constant specifying how to deal with file-level conflicts.
            For all but NORMAL, the index will not record a conflict.

        flags
            A combination of enums.MergeFlag constants.

        file_flags
            A combination of enums.MergeFileFlag constants.
        """

        if isinstance(source, Reference):
            # Annotated commit from ref
            cptr = ffi.new('struct git_reference **')
            ffi.buffer(cptr)[:] = source._pointer[:]  # type: ignore[attr-defined]
            commit_ptr = ffi.new('git_annotated_commit **')
            err = C.git_annotated_commit_from_ref(commit_ptr, self._repo, cptr[0])
            check_error(err)
        else:
            # Annotated commit from commit id
            if isinstance(source, str):
                # For backwards compatibility, parse a string as a partial commit hash
                warnings.warn(
                    'Passing str to Repository.merge is deprecated. '
                    'Pass Commit, Oid, or a Reference (such as a Branch) instead.',
                    DeprecationWarning,
                )
                oid = self[source].peel(Commit).id
            elif isinstance(source, Commit):
                oid = source.id
            elif isinstance(source, Oid):
                oid = source
            else:
                raise TypeError('expected Reference, Commit, or Oid')
            c_id = ffi.new('git_oid *')
            ffi.buffer(c_id)[:] = oid.raw[:]
            commit_ptr = ffi.new('git_annotated_commit **')
            err = C.git_annotated_commit_lookup(commit_ptr, self._repo, c_id)
            check_error(err)

        merge_opts = self._merge_options(favor, flags, file_flags)

        checkout_opts = ffi.new('git_checkout_options *')
        C.git_checkout_options_init(checkout_opts, 1)
        checkout_opts.checkout_strategy = int(
            CheckoutStrategy.SAFE | CheckoutStrategy.RECREATE_MISSING
        )

        err = C.git_merge(self._repo, commit_ptr, 1, merge_opts, checkout_opts)
        C.git_annotated_commit_free(commit_ptr[0])
        check_error(err)

    #
    # Prepared message (MERGE_MSG)
    #
    @property
    def raw_message(self) -> bytes:
        """
        Retrieve git's prepared message (bytes).
        See `Repository.message` for more information.
        """
        buf = ffi.new('git_buf *', (ffi.NULL, 0))
        try:
            err = C.git_repository_message(buf, self._repo)
            if err == C.GIT_ENOTFOUND:
                return b''
            check_error(err)
            return ffi.string(buf.ptr)
        finally:
            C.git_buf_dispose(buf)

    @property
    def message(self) -> str:
        """
        Retrieve git's prepared message.

        Operations such as git revert/cherry-pick/merge with the -n option stop
        just short of creating a commit with the changes and save their
        prepared message in .git/MERGE_MSG so the next git-commit execution can
        present it to the user for them to amend if they wish.

        Use this function to get the contents of this file. Don't forget to
        call `Repository.remove_message()` after you create the commit.

        Note that the message is also removed by `Repository.state_cleanup()`.

        If there is no such message, an empty string is returned.
        """
        return self.raw_message.decode('utf-8')

    def remove_message(self) -> None:
        """
        Remove git's prepared message.
        """
        err = C.git_repository_message_remove(self._repo)
        check_error(err)

    #
    # Describe
    #
    def describe(
        self,
        committish: str | Reference | Commit | None = None,
        max_candidates_tags: int | None = None,
        describe_strategy: DescribeStrategy = DescribeStrategy.DEFAULT,
        pattern: str | None = None,
        only_follow_first_parent: bool | None = None,
        show_commit_oid_as_fallback: bool | None = None,
        abbreviated_size: int | None = None,
        always_use_long_format: bool | None = None,
        dirty_suffix: str | None = None,
    ) -> str:
        """
        Describe a commit-ish or the current working tree.

        Returns: The description (str).

        Parameters:

        committish : `str`, :class:`~.Reference`, or :class:`~.Commit`
            Commit-ish object or object name to describe, or `None` to describe
            the current working tree.

        max_candidates_tags : int
            The number of candidate tags to consider. Increasing above 10 will
            take slightly longer but may produce a more accurate result. A
            value of 0 will cause only exact matches to be output.

        describe_strategy : DescribeStrategy
            Can be one of:

            * `DescribeStrategy.DEFAULT` - Only match annotated tags.
            * `DescribeStrategy.TAGS` - Match everything under refs/tags/
              (includes lightweight tags).
            * `DescribeStrategy.ALL` - Match everything under refs/ (includes
              branches).

        pattern : str
            Only consider tags matching the given `glob(7)` pattern, excluding
            the "refs/tags/" prefix.

        only_follow_first_parent : bool
            Follow only the first parent commit upon seeing a merge commit.

        show_commit_oid_as_fallback : bool
            Show uniquely abbreviated commit object as fallback.

        abbreviated_size : int
            The minimum number of hexadecimal digits to show for abbreviated
            object names. A value of 0 will suppress long format, only showing
            the closest tag.

        always_use_long_format : bool
            Always output the long format (the nearest tag, the number of
            commits, and the abbreviated commit name) even when the committish
            matches a tag.

        dirty_suffix : str
            A string to append if the working tree is dirty.

        Example::

            repo.describe(pattern='public/*', dirty_suffix='-dirty')
        """

        options = ffi.new('git_describe_options *')
        C.git_describe_options_init(options, C.GIT_DESCRIBE_OPTIONS_VERSION)

        if max_candidates_tags is not None:
            options.max_candidates_tags = max_candidates_tags
        if describe_strategy is not None:
            options.describe_strategy = int(describe_strategy)
        if pattern:
            # The returned pointer object has ownership on the allocated
            # memory. Make sure it is kept alive until git_describe_commit() or
            # git_describe_workdir() are called below.
            pattern_char = ffi.new('char[]', to_bytes(pattern))
            options.pattern = pattern_char
        if only_follow_first_parent is not None:
            options.only_follow_first_parent = only_follow_first_parent
        if show_commit_oid_as_fallback is not None:
            options.show_commit_oid_as_fallback = show_commit_oid_as_fallback

        result = ffi.new('git_describe_result **')
        if committish:
            committish_rev: Object | Reference | Commit
            if isinstance(committish, str):
                committish_rev = self.revparse_single(committish)
            else:
                committish_rev = committish

            commit = committish_rev.peel(Commit)

            cptr = ffi.new('git_object **')
            ffi.buffer(cptr)[:] = commit._pointer[:]

            err = C.git_describe_commit(result, cptr[0], options)
        else:
            err = C.git_describe_workdir(result, self._repo, options)
        check_error(err)

        try:
            format_options = ffi.new('git_describe_format_options *')
            C.git_describe_init_format_options(
                format_options, C.GIT_DESCRIBE_FORMAT_OPTIONS_VERSION
            )

            if abbreviated_size is not None:
                format_options.abbreviated_size = abbreviated_size
            if always_use_long_format is not None:
                format_options.always_use_long_format = always_use_long_format
            dirty_ptr = None
            if dirty_suffix:
                dirty_ptr = ffi.new('char[]', to_bytes(dirty_suffix))
                format_options.dirty_suffix = dirty_ptr

            buf = ffi.new('git_buf *', (ffi.NULL, 0))

            err = C.git_describe_format(buf, result[0], format_options)
            check_error(err)

            try:
                return ffi.string(buf.ptr).decode('utf-8')
            finally:
                C.git_buf_dispose(buf)
        finally:
            C.git_describe_result_free(result[0])

    #
    # Stash
    #
    def stash(
        self,
        stasher: Signature,
        message: str | None = None,
        keep_index: bool = False,
        include_untracked: bool = False,
        include_ignored: bool = False,
        keep_all: bool = False,
        paths: list[str] | None = None,
    ) -> Oid:
        """
        Save changes to the working directory to the stash.

        Returns: The Oid of the stash merge commit (Oid).

        Parameters:

        stasher : Signature
            The identity of the person doing the stashing.

        message : str
            An optional description of stashed state.

        keep_index : bool
            Leave changes already added to the index in the working directory.

        include_untracked : bool
            Also stash untracked files.

        include_ignored : bool
            Also stash ignored files.

        keep_all : bool
            All changes in the index and working directory are left intact.

        paths : list[str]
            An optional list of paths that control which files are stashed.

        Example::

            >>> repo = pygit2.Repository('.')
            >>> repo.stash(repo.default_signature(), 'WIP: stashing')
        """

        opts = ffi.new('git_stash_save_options *')
        C.git_stash_save_options_init(opts, C.GIT_STASH_SAVE_OPTIONS_VERSION)

        flags = 0
        flags |= keep_index * C.GIT_STASH_KEEP_INDEX
        flags |= keep_all * C.GIT_STASH_KEEP_ALL
        flags |= include_untracked * C.GIT_STASH_INCLUDE_UNTRACKED
        flags |= include_ignored * C.GIT_STASH_INCLUDE_IGNORED
        opts.flags = flags

        stasher_cptr = ffi.new('git_signature **')
        ffi.buffer(stasher_cptr)[:] = stasher._pointer[:]
        opts.stasher = stasher_cptr[0]

        if message:
            message_ref = ffi.new('char[]', to_bytes(message))
            opts.message = message_ref

        if paths:
            arr = StrArray(paths)
            opts.paths = arr.ptr[0]  # type: ignore[index]

        coid = ffi.new('git_oid *')
        err = C.git_stash_save_with_opts(coid, self._repo, opts)

        check_error(err)

        return Oid(raw=bytes(ffi.buffer(coid)[:]))

    def stash_apply(
        self,
        index: int = 0,
        reinstate_index: bool = False,
        strategy: CheckoutStrategy | None = None,
        callbacks: StashApplyCallbacks | None = None,
    ) -> None:
        """
        Apply a stashed state in the stash list to the working directory.

        Parameters:

        index : int
            The position within the stash list of the stash to apply. 0 is the
            most recent stash.

        reinstate_index : bool
            Try to reinstate stashed changes to the index.

        callbacks : StashApplyCallbacks
            Optional. Supply a `callbacks` object to get information about
            the progress of the stash application as it is being performed.

            The callbacks should be an object which inherits from
            `pyclass:StashApplyCallbacks`. It should implement the callbacks
            as overridden methods.

            Note that this class inherits from CheckoutCallbacks, so you can
            also get information from the checkout part of the unstashing
            process via the callbacks.

        The checkout options may be customized using the same arguments taken by
        Repository.checkout().

        Example::

            >>> repo = pygit2.Repository('.')
            >>> repo.stash(repo.default_signature(), 'WIP: stashing')
            >>> repo.stash_apply(strategy=CheckoutStrategy.ALLOW_CONFLICTS)
        """
        with git_stash_apply_options(
            reinstate_index=reinstate_index,
            strategy=strategy,
            callbacks=callbacks,
        ) as payload:
            err = C.git_stash_apply(self._repo, index, payload.stash_apply_options)
            payload.check_error(err)

    def stash_drop(self, index: int = 0) -> None:
        """
        Remove a stashed state from the stash list.

        Parameters:

        index : int
            The position within the stash list of the stash to remove. 0 is
            the most recent stash.
        """
        check_error(C.git_stash_drop(self._repo, index))

    def stash_pop(
        self,
        index: int = 0,
        reinstate_index: bool = False,
        strategy: CheckoutStrategy | None = None,
        callbacks: StashApplyCallbacks | None = None,
    ) -> None:
        """Apply a stashed state and remove it from the stash list.

        For arguments, see Repository.stash_apply().
        """
        with git_stash_apply_options(
            reinstate_index=reinstate_index,
            strategy=strategy,
            callbacks=callbacks,
        ) as payload:
            err = C.git_stash_pop(self._repo, index, payload.stash_apply_options)
            payload.check_error(err)

    #
    # Utility for writing a tree into an archive
    #
    def write_archive(
        self,
        treeish: str | Tree | Object | Oid,
        archive: tarfile.TarFile,
        timestamp: int | None = None,
        prefix: str = '',
    ) -> None:
        """
        Write treeish into an archive.

        If no timestamp is provided and 'treeish' is a commit, its committer
        timestamp will be used. Otherwise the current time will be used.

        All path names in the archive are added to 'prefix', which defaults to
        an empty string.

        Parameters:

        treeish
            The treeish to write.

        archive
            An archive from the 'tarfile' module.

        timestamp
            Timestamp to use for the files in the archive.

        prefix
            Extra prefix to add to the path names in the archive.

        Example::

            >>> import tarfile, pygit2
            >>> with tarfile.open('foo.tar', 'w') as archive:
            >>>     repo = pygit2.Repository('.')
            >>>     repo.write_archive(repo.head.target, archive)
        """

        # Try to get a tree form whatever we got
        if isinstance(treeish, (str, Oid)):
            treeish = self[treeish]

        tree = treeish.peel(Tree)

        # if we don't have a timestamp, try to get it from a commit
        if not timestamp:
            try:
                commit = treeish.peel(Commit)
                timestamp = commit.committer.time
            except Exception:
                pass

        # as a last resort, use the current timestamp
        if not timestamp:
            timestamp = int(time())

        index = Index()
        index.read_tree(tree)

        for entry in index:
            content = self[entry.id].read_raw()
            info = tarfile.TarInfo(prefix + entry.path)
            info.size = len(content)
            info.mtime = timestamp
            info.uname = info.gname = 'root'  # just because git does this
            if entry.mode == FileMode.LINK:
                info.type = tarfile.SYMTYPE
                info.linkname = content.decode('utf-8')
                info.mode = 0o777  # symlinks get placeholder
                info.size = 0
                archive.addfile(info)
            else:
                info.mode = entry.mode
                archive.addfile(info, BytesIO(content))

    #
    # Ahead-behind, which mostly lives on its own namespace
    #
    def ahead_behind(self, local: Oid | str, upstream: Oid | str) -> tuple[int, int]:
        """
        Calculate how many different commits are in the non-common parts of the
        history between the two given ids.

        Ahead is how many commits are in the ancestry of the `local` commit
        which are not in the `upstream` commit. Behind is the opposite.

        Returns: a tuple of two integers with the number of commits ahead and
        behind respectively.

        Parameters:

        local
            The commit which is considered the local or current state.

        upstream
            The commit which is considered the upstream.
        """

        if not isinstance(local, Oid):
            local = self.expand_id(local)

        if not isinstance(upstream, Oid):
            upstream = self.expand_id(upstream)

        ahead, behind = ffi.new('size_t*'), ffi.new('size_t*')
        oid1, oid2 = ffi.new('git_oid *'), ffi.new('git_oid *')
        ffi.buffer(oid1)[:] = local.raw[:]
        ffi.buffer(oid2)[:] = upstream.raw[:]
        err = C.git_graph_ahead_behind(ahead, behind, self._repo, oid1, oid2)
        check_error(err)

        return int(ahead[0]), int(behind[0])

    #
    # Git attributes
    #
    def get_attr(
        self,
        path: str | bytes | Path,
        name: str | bytes,
        flags: AttrCheck = AttrCheck.FILE_THEN_INDEX,
        commit: Oid | str | None = None,
    ) -> bool | None | str:
        """
        Retrieve an attribute for a file by path.

        Returns: a boolean, `None` if the value is unspecified, or string with
        the value of the attribute.

        Parameters:

        path
            The path of the file to look up attributes for, relative to the
            workdir root.

        name
            The name of the attribute to look up.

        flags
            A combination of enums.AttrCheck flags which determine the lookup order.

        commit
            Optional id of commit to load attributes from when the
            `INCLUDE_COMMIT` flag is specified.

        Examples::

            >>> print(repo.get_attr('splash.bmp', 'binary'))
            True
            >>> print(repo.get_attr('splash.bmp', 'unknown-attr'))
            None
            >>> repo.get_attr('test.h', 'whitespace')
            'tab-in-indent,trailing-space'
        """

        copts = ffi.new('git_attr_options *')
        copts.version = C.GIT_ATTR_OPTIONS_VERSION
        copts.flags = int(flags)
        if commit is not None:
            if not isinstance(commit, Oid):
                commit = Oid(hex=commit)
            ffi.buffer(ffi.addressof(copts, 'attr_commit_id'))[:] = commit.raw

        cvalue = ffi.new('char **')
        err = C.git_attr_get_ext(
            cvalue, self._repo, copts, to_bytes(path), to_bytes(name)
        )
        check_error(err)

        # Now let's see if we can figure out what the value is
        attr_kind = C.git_attr_value(cvalue[0])
        if attr_kind == C.GIT_ATTR_VALUE_UNSPECIFIED:
            return None
        elif attr_kind == C.GIT_ATTR_VALUE_TRUE:
            return True
        elif attr_kind == C.GIT_ATTR_VALUE_FALSE:
            return False
        elif attr_kind == C.GIT_ATTR_VALUE_STRING:
            return ffi.string(cvalue[0]).decode('utf-8')

        assert False, 'the attribute value from libgit2 is invalid'

    #
    # Identity for reference operations
    #
    @property
    def ident(self):
        cname = ffi.new('char **')
        cemail = ffi.new('char **')

        err = C.git_repository_ident(cname, cemail, self._repo)
        check_error(err)

        return (ffi.string(cname).decode('utf-8'), ffi.string(cemail).decode('utf-8'))

    def set_ident(self, name: str, email: str) -> None:
        """Set the identity to be used for reference operations.

        Updates to some references also append data to their
        reflog. You can use this method to set what identity will be
        used. If none is set, it will be read from the configuration.
        """

        err = C.git_repository_set_ident(self._repo, to_bytes(name), to_bytes(email))
        check_error(err)

    def revert(self, commit: Commit) -> None:
        """
        Revert the given commit, producing changes in the index and working
        directory.

        This operation updates the repository's state and prepared message
        (MERGE_MSG).
        """
        commit_ptr = ffi.new('git_commit **')
        ffi.buffer(commit_ptr)[:] = commit._pointer[:]
        err = C.git_revert(self._repo, commit_ptr[0], ffi.NULL)
        check_error(err)

    def revert_commit(
        self, revert_commit: Commit, our_commit: Commit, mainline: int = 0
    ) -> Index:
        """
        Revert the given Commit against the given "our" Commit, producing an
        Index that reflects the result of the revert.

        Returns: an Index with the result of the revert.

        Parameters:

        revert_commit
            The Commit to revert.

        our_commit
            The Commit to revert against (eg, HEAD).

        mainline
            The parent of the revert Commit, if it is a merge (i.e. 1, 2).
        """
        cindex = ffi.new('git_index **')
        revert_commit_ptr = ffi.new('git_commit **')
        our_commit_ptr = ffi.new('git_commit **')

        ffi.buffer(revert_commit_ptr)[:] = revert_commit._pointer[:]
        ffi.buffer(our_commit_ptr)[:] = our_commit._pointer[:]

        opts = ffi.new('git_merge_options *')
        err = C.git_merge_options_init(opts, C.GIT_MERGE_OPTIONS_VERSION)
        check_error(err)

        err = C.git_revert_commit(
            cindex, self._repo, revert_commit_ptr[0], our_commit_ptr[0], mainline, opts
        )
        check_error(err)

        return Index.from_c(self, cindex)

    #
    # Amend commit
    #
    def amend_commit(
        self,
        commit: Commit | Oid | str,
        refname: Reference | str | None,
        author: Signature | None = None,
        committer: Signature | None = None,
        message: str | None = None,
        tree: Tree | Oid | str | None = None,
        encoding: str = 'UTF-8',
    ) -> Oid:
        """
        Amend an existing commit by replacing only explicitly passed values,
        return the rewritten commit's oid.

        This creates a new commit that is exactly the same as the old commit,
        except that any explicitly passed values will be updated. The new
        commit has the same parents as the old commit.

        You may omit the `author`, `committer`, `message`, `tree`, and
        `encoding` parameters, in which case this will use the values
        from the original `commit`.

        Parameters:

        commit : Commit, Oid, or str
            The commit to amend.

        refname : Reference or str
            If not `None`, name of the reference that will be updated to point
            to the newly rewritten commit. Use "HEAD" to update the HEAD of the
            current branch and make it point to the rewritten commit.
            If you want to amend a commit that is not currently the tip of the
            branch and then rewrite the following commits to reach a ref, pass
            this as `None` and update the rest of the commit chain and ref
            separately.

        author : Signature
            If not None, replace the old commit's author signature with this
            one.

        committer : Signature
            If not None, replace the old commit's committer signature with this
            one.

        message : str
            If not None, replace the old commit's message with this one.

        tree : Tree, Oid, or str
            If not None, replace the old commit's tree with this one.

        encoding : str
            Optional encoding for `message`.
        """

        # Initialize parameters to pass on to C function git_commit_amend.
        # Note: the pointers are all initialized to NULL by default.
        coid = ffi.new('git_oid *')
        commit_cptr = ffi.new('git_commit **')
        refname_cstr: 'ArrayC[char]' | 'ffi.NULL_TYPE' = ffi.NULL
        author_cptr = ffi.new('git_signature **')
        committer_cptr = ffi.new('git_signature **')
        message_cstr: 'ArrayC[char]' | 'ffi.NULL_TYPE' = ffi.NULL
        encoding_cstr: 'ArrayC[char]' | 'ffi.NULL_TYPE' = ffi.NULL
        tree_cptr = ffi.new('git_tree **')

        # Get commit as pointer to git_commit.
        if isinstance(commit, (str, Oid)):
            commit_object = self[commit]
            commit_commit = commit_object.peel(Commit)
        elif isinstance(commit, Commit):
            commit_commit = commit
        elif commit is None:
            raise ValueError('the commit to amend cannot be None')
        else:
            raise TypeError('the commit to amend must be a Commit, str, or Oid')
        ffi.buffer(commit_cptr)[:] = commit_commit._pointer[:]

        # Get refname as C string.
        if isinstance(refname, Reference):
            refname_cstr = ffi.new('char[]', to_bytes(refname.name))
        elif type(refname) is str:
            refname_cstr = ffi.new('char[]', to_bytes(refname))
        elif refname is not None:
            raise TypeError('refname must be a str or Reference')

        # Get author as pointer to git_signature.
        if isinstance(author, Signature):
            ffi.buffer(author_cptr)[:] = author._pointer[:]
        elif author is not None:
            raise TypeError('author must be a Signature')

        # Get committer as pointer to git_signature.
        if isinstance(committer, Signature):
            ffi.buffer(committer_cptr)[:] = committer._pointer[:]
        elif committer is not None:
            raise TypeError('committer must be a Signature')

        # Get message and encoding as C strings.
        if message is not None:
            message_cstr = ffi.new('char[]', to_bytes(message, encoding))
            encoding_cstr = ffi.new('char[]', to_bytes(encoding))

        # Get tree as pointer to git_tree.
        if tree is not None:
            if isinstance(tree, (str, Oid)):
                tree_object = self[tree]
            else:
                tree_object = tree
            tree_tree = tree_object.peel(Tree)
            ffi.buffer(tree_cptr)[:] = tree_tree._pointer[:]

        # Amend the commit.
        err = C.git_commit_amend(
            coid,
            commit_cptr[0],
            refname_cstr,
            author_cptr[0],
            committer_cptr[0],
            encoding_cstr,
            message_cstr,
            tree_cptr[0],
        )
        check_error(err)

        return Oid(raw=bytes(ffi.buffer(coid)[:]))

    def __ensure_tree(self, maybe_tree: str | Oid | Tree) -> Tree:
        if isinstance(maybe_tree, Tree):
            return maybe_tree
        return self[maybe_tree].peel(Tree)


class Repository(BaseRepository):
    def __init__(
        self,
        path: str | bytes | None | Path = None,
        flags: RepositoryOpenFlag = RepositoryOpenFlag.DEFAULT,
    ):
        """
        The Repository constructor will commonly be called with one argument,
        the path of the repository to open.

        Alternatively, constructing a repository with no arguments will create
        a repository with no backends. You can use this path to create
        repositories with custom backends. Note that most operations on the
        repository are considered invalid and may lead to undefined behavior if
        attempted before providing an odb and refdb via set_odb and set_refdb.

        Parameters:

        path : str
        The path to open - if not provided, the repository will have no backend.

        flags : enums.RepositoryOpenFlag
        An optional combination of enums.RepositoryOpenFlag constants
        controlling how to open the repository.
        """

        if path is not None:
            if hasattr(path, '__fspath__'):
                path = path.__fspath__()
            if not isinstance(path, str):
                path = path.decode('utf-8')
            path_backend = init_file_backend(path, int(flags))
            super().__init__(path_backend)
        else:
            super().__init__()

    @classmethod
    def _from_c(cls, ptr: 'GitRepositoryC', owned: bool) -> 'Repository':
        cptr = ffi.new('git_repository **')
        cptr[0] = ptr
        repo = cls.__new__(cls)
        BaseRepository._from_c(repo, bytes(ffi.buffer(cptr)[:]), owned)  # type: ignore
        repo._common_init()
        return repo



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING

from ._pygit2 import Branch, Commit, Oid
from .enums import BranchType, ReferenceType

# Need BaseRepository for type hints, but don't let it cause a circular dependency
if TYPE_CHECKING:
    from .repository import BaseRepository


class Branches:
    local: 'Branches'
    remote: 'Branches'

    def __init__(
        self,
        repository: BaseRepository,
        flag: BranchType = BranchType.ALL,
        commit: Commit | Oid | str | None = None,
    ) -> None:
        self._repository = repository
        self._flag = flag
        if commit is not None:
            if isinstance(commit, Commit):
                commit = commit.id
            elif not isinstance(commit, Oid):
                commit = self._repository.expand_id(commit)
        self._commit = commit

        if flag == BranchType.ALL:
            self.local = Branches(repository, flag=BranchType.LOCAL, commit=commit)
            self.remote = Branches(repository, flag=BranchType.REMOTE, commit=commit)

    def __getitem__(self, name: str) -> Branch:
        branch = None
        if self._flag & BranchType.LOCAL:
            branch = self._repository.lookup_branch(name, BranchType.LOCAL)

        if branch is None and self._flag & BranchType.REMOTE:
            branch = self._repository.lookup_branch(name, BranchType.REMOTE)

        if branch is None or not self._valid(branch):
            raise KeyError(f'Branch not found: {name}')

        return branch

    def get(self, key: str) -> Branch:
        try:
            return self[key]
        except KeyError:
            return None  # type:ignore #  next commit

    def __iter__(self) -> Iterator[str]:
        for branch_name in self._repository.listall_branches(self._flag):
            if self._commit is None or self.get(branch_name) is not None:
                yield branch_name

    def create(self, name: str, commit: Commit, force: bool = False) -> Branch:
        return self._repository.create_branch(name, commit, force)

    def delete(self, name: str) -> None:
        self[name].delete()

    def _valid(self, branch: Branch) -> bool:
        if branch.type == ReferenceType.SYMBOLIC:
            branch_direct = branch.resolve()
        else:
            branch_direct = branch

        return (
            self._commit is None
            or branch_direct.target == self._commit
            or self._repository.descendant_of(branch_direct.target, self._commit)
        )

    def with_commit(self, commit: Commit | Oid | str | None) -> 'Branches':
        assert self._commit is None
        return Branches(self._repository, self._flag, commit)

    def __contains__(self, name: str) -> bool:
        return self.get(name) is not None



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""
Libgit2 global options management using CFFI.
"""

from __future__ import annotations

# Import only for type checking to avoid circular imports
from typing import TYPE_CHECKING, Any, Literal, cast, overload

from .errors import check_error
from .ffi import C, ffi
from .utils import to_bytes, to_str

if TYPE_CHECKING:
    from ._libgit2.ffi import NULL_TYPE, ArrayC, char, char_pointer
    from .enums import ConfigLevel, ObjectType, Option

# Export GIT_OPT constants for backward compatibility
GIT_OPT_GET_MWINDOW_SIZE: int = C.GIT_OPT_GET_MWINDOW_SIZE
GIT_OPT_SET_MWINDOW_SIZE: int = C.GIT_OPT_SET_MWINDOW_SIZE
GIT_OPT_GET_MWINDOW_MAPPED_LIMIT: int = C.GIT_OPT_GET_MWINDOW_MAPPED_LIMIT
GIT_OPT_SET_MWINDOW_MAPPED_LIMIT: int = C.GIT_OPT_SET_MWINDOW_MAPPED_LIMIT
GIT_OPT_GET_SEARCH_PATH: int = C.GIT_OPT_GET_SEARCH_PATH
GIT_OPT_SET_SEARCH_PATH: int = C.GIT_OPT_SET_SEARCH_PATH
GIT_OPT_SET_CACHE_OBJECT_LIMIT: int = C.GIT_OPT_SET_CACHE_OBJECT_LIMIT
GIT_OPT_SET_CACHE_MAX_SIZE: int = C.GIT_OPT_SET_CACHE_MAX_SIZE
GIT_OPT_ENABLE_CACHING: int = C.GIT_OPT_ENABLE_CACHING
GIT_OPT_GET_CACHED_MEMORY: int = C.GIT_OPT_GET_CACHED_MEMORY
GIT_OPT_GET_TEMPLATE_PATH: int = C.GIT_OPT_GET_TEMPLATE_PATH
GIT_OPT_SET_TEMPLATE_PATH: int = C.GIT_OPT_SET_TEMPLATE_PATH
GIT_OPT_SET_SSL_CERT_LOCATIONS: int = C.GIT_OPT_SET_SSL_CERT_LOCATIONS
GIT_OPT_SET_USER_AGENT: int = C.GIT_OPT_SET_USER_AGENT
GIT_OPT_ENABLE_STRICT_OBJECT_CREATION: int = C.GIT_OPT_ENABLE_STRICT_OBJECT_CREATION
GIT_OPT_ENABLE_STRICT_SYMBOLIC_REF_CREATION: int = (
    C.GIT_OPT_ENABLE_STRICT_SYMBOLIC_REF_CREATION
)
GIT_OPT_SET_SSL_CIPHERS: int = C.GIT_OPT_SET_SSL_CIPHERS
GIT_OPT_GET_USER_AGENT: int = C.GIT_OPT_GET_USER_AGENT
GIT_OPT_ENABLE_OFS_DELTA: int = C.GIT_OPT_ENABLE_OFS_DELTA
GIT_OPT_ENABLE_FSYNC_GITDIR: int = C.GIT_OPT_ENABLE_FSYNC_GITDIR
GIT_OPT_GET_WINDOWS_SHAREMODE: int = C.GIT_OPT_GET_WINDOWS_SHAREMODE
GIT_OPT_SET_WINDOWS_SHAREMODE: int = C.GIT_OPT_SET_WINDOWS_SHAREMODE
GIT_OPT_ENABLE_STRICT_HASH_VERIFICATION: int = C.GIT_OPT_ENABLE_STRICT_HASH_VERIFICATION
GIT_OPT_SET_ALLOCATOR: int = C.GIT_OPT_SET_ALLOCATOR
GIT_OPT_ENABLE_UNSAVED_INDEX_SAFETY: int = C.GIT_OPT_ENABLE_UNSAVED_INDEX_SAFETY
GIT_OPT_GET_PACK_MAX_OBJECTS: int = C.GIT_OPT_GET_PACK_MAX_OBJECTS
GIT_OPT_SET_PACK_MAX_OBJECTS: int = C.GIT_OPT_SET_PACK_MAX_OBJECTS
GIT_OPT_DISABLE_PACK_KEEP_FILE_CHECKS: int = C.GIT_OPT_DISABLE_PACK_KEEP_FILE_CHECKS
GIT_OPT_GET_MWINDOW_FILE_LIMIT: int = C.GIT_OPT_GET_MWINDOW_FILE_LIMIT
GIT_OPT_SET_MWINDOW_FILE_LIMIT: int = C.GIT_OPT_SET_MWINDOW_FILE_LIMIT
GIT_OPT_GET_OWNER_VALIDATION: int = C.GIT_OPT_GET_OWNER_VALIDATION
GIT_OPT_SET_OWNER_VALIDATION: int = C.GIT_OPT_SET_OWNER_VALIDATION
GIT_OPT_ENABLE_HTTP_EXPECT_CONTINUE: int = C.GIT_OPT_ENABLE_HTTP_EXPECT_CONTINUE
GIT_OPT_SET_ODB_PACKED_PRIORITY: int = C.GIT_OPT_SET_ODB_PACKED_PRIORITY
GIT_OPT_SET_ODB_LOOSE_PRIORITY: int = C.GIT_OPT_SET_ODB_LOOSE_PRIORITY
GIT_OPT_GET_EXTENSIONS: int = C.GIT_OPT_GET_EXTENSIONS
GIT_OPT_SET_EXTENSIONS: int = C.GIT_OPT_SET_EXTENSIONS
GIT_OPT_GET_HOMEDIR: int = C.GIT_OPT_GET_HOMEDIR
GIT_OPT_SET_HOMEDIR: int = C.GIT_OPT_SET_HOMEDIR
GIT_OPT_SET_SERVER_CONNECT_TIMEOUT: int = C.GIT_OPT_SET_SERVER_CONNECT_TIMEOUT
GIT_OPT_GET_SERVER_CONNECT_TIMEOUT: int = C.GIT_OPT_GET_SERVER_CONNECT_TIMEOUT
GIT_OPT_SET_SERVER_TIMEOUT: int = C.GIT_OPT_SET_SERVER_TIMEOUT
GIT_OPT_GET_SERVER_TIMEOUT: int = C.GIT_OPT_GET_SERVER_TIMEOUT
GIT_OPT_GET_USER_AGENT_PRODUCT: int = C.GIT_OPT_GET_USER_AGENT_PRODUCT
GIT_OPT_SET_USER_AGENT_PRODUCT: int = C.GIT_OPT_SET_USER_AGENT_PRODUCT
GIT_OPT_ADD_SSL_X509_CERT: int = C.GIT_OPT_ADD_SSL_X509_CERT


NOT_PASSED = object()


def check_args(option: Option, arg1: Any, arg2: Any, expected: int) -> None:
    if expected == 0 and (arg1 is not NOT_PASSED or arg2 is not NOT_PASSED):
        raise TypeError(f'option({option}) takes no additional arguments')

    if expected == 1 and (arg1 is NOT_PASSED or arg2 is not NOT_PASSED):
        raise TypeError(f'option({option}, x) requires 1 additional argument')

    if expected == 2 and (arg1 is NOT_PASSED or arg2 is NOT_PASSED):
        raise TypeError(f'option({option}, x, y) requires 2 additional arguments')


@overload
def option(
    option_type: Literal[
        Option.GET_MWINDOW_SIZE,
        Option.GET_MWINDOW_MAPPED_LIMIT,
        Option.GET_MWINDOW_FILE_LIMIT,
    ],
) -> int: ...


@overload
def option(
    option_type: Literal[
        Option.SET_MWINDOW_SIZE,
        Option.SET_MWINDOW_MAPPED_LIMIT,
        Option.SET_MWINDOW_FILE_LIMIT,
        Option.SET_CACHE_MAX_SIZE,
    ],
    arg1: int,  # value
) -> None: ...


@overload
def option(
    option_type: Literal[Option.GET_SEARCH_PATH],
    arg1: ConfigLevel,  # value
) -> str: ...


@overload
def option(
    option_type: Literal[Option.SET_SEARCH_PATH],
    arg1: ConfigLevel,  # type
    arg2: str,  # value
) -> None: ...


@overload
def option(
    option_type: Literal[Option.SET_CACHE_OBJECT_LIMIT],
    arg1: ObjectType,  # type
    arg2: int,  # limit
) -> None: ...


@overload
def option(option_type: Literal[Option.GET_CACHED_MEMORY]) -> tuple[int, int]: ...


@overload
def option(
    option_type: Literal[Option.SET_SSL_CERT_LOCATIONS],
    arg1: str | bytes | None,  # cert_file
    arg2: str | bytes | None,  # cert_dir
) -> None: ...


@overload
def option(
    option_type: Literal[
        Option.ENABLE_CACHING,
        Option.ENABLE_STRICT_OBJECT_CREATION,
        Option.ENABLE_STRICT_SYMBOLIC_REF_CREATION,
        Option.ENABLE_OFS_DELTA,
        Option.ENABLE_FSYNC_GITDIR,
        Option.ENABLE_STRICT_HASH_VERIFICATION,
        Option.ENABLE_UNSAVED_INDEX_SAFETY,
        Option.DISABLE_PACK_KEEP_FILE_CHECKS,
        Option.SET_OWNER_VALIDATION,
    ],
    arg1: bool,  # value
) -> None: ...


@overload
def option(option_type: Literal[Option.GET_OWNER_VALIDATION]) -> bool: ...


@overload
def option(
    option_type: Literal[
        Option.GET_TEMPLATE_PATH,
        Option.GET_USER_AGENT,
        Option.GET_HOMEDIR,
        Option.GET_USER_AGENT_PRODUCT,
    ],
) -> str | None: ...


@overload
def option(
    option_type: Literal[
        Option.SET_TEMPLATE_PATH,
        Option.SET_USER_AGENT,
        Option.SET_SSL_CIPHERS,
        Option.SET_HOMEDIR,
        Option.SET_USER_AGENT_PRODUCT,
    ],
    arg1: str | bytes,  # value
) -> None: ...


@overload
def option(
    option_type: Literal[
        Option.GET_WINDOWS_SHAREMODE,
        Option.GET_PACK_MAX_OBJECTS,
        Option.GET_SERVER_CONNECT_TIMEOUT,
        Option.GET_SERVER_TIMEOUT,
    ],
) -> int: ...


@overload
def option(
    option_type: Literal[
        Option.SET_WINDOWS_SHAREMODE,
        Option.SET_PACK_MAX_OBJECTS,
        Option.ENABLE_HTTP_EXPECT_CONTINUE,
        Option.SET_ODB_PACKED_PRIORITY,
        Option.SET_ODB_LOOSE_PRIORITY,
        Option.SET_SERVER_CONNECT_TIMEOUT,
        Option.SET_SERVER_TIMEOUT,
    ],
    arg1: int,  # value
) -> None: ...


@overload
def option(option_type: Literal[Option.GET_EXTENSIONS]) -> list[str]: ...


@overload
def option(
    option_type: Literal[Option.SET_EXTENSIONS],
    arg1: list[str],  # extensions
    arg2: int,  # length
) -> None: ...


@overload
def option(
    option_type: Literal[Option.ADD_SSL_X509_CERT],
    arg1: str | bytes,  # certificate
) -> None: ...


# Fallback overload for generic Option values (used in tests)
@overload
def option(option_type: Option, arg1: Any = ..., arg2: Any = ...) -> Any: ...


def option(option_type: Option, arg1: Any = NOT_PASSED, arg2: Any = NOT_PASSED) -> Any:
    """
    Get or set a libgit2 option.

    Parameters:

    GIT_OPT_GET_SEARCH_PATH, level
        Get the config search path for the given level.

    GIT_OPT_SET_SEARCH_PATH, level, path
        Set the config search path for the given level.

    GIT_OPT_GET_MWINDOW_SIZE
        Get the maximum mmap window size.

    GIT_OPT_SET_MWINDOW_SIZE, size
        Set the maximum mmap window size.

    GIT_OPT_GET_MWINDOW_FILE_LIMIT
        Get the maximum number of files that will be mapped at any time by the library.

    GIT_OPT_SET_MWINDOW_FILE_LIMIT, size
        Set the maximum number of files that can be mapped at any time by the library. The default (0) is unlimited.

    GIT_OPT_GET_OWNER_VALIDATION
        Gets the owner validation setting for repository directories.

    GIT_OPT_SET_OWNER_VALIDATION, enabled
        Set that repository directories should be owned by the current user.
        The default is to validate ownership.

    GIT_OPT_GET_TEMPLATE_PATH
        Get the default template path.

    GIT_OPT_SET_TEMPLATE_PATH, path
        Set the default template path.

    GIT_OPT_GET_USER_AGENT
        Get the user agent string.

    GIT_OPT_SET_USER_AGENT, user_agent
        Set the user agent string.

    GIT_OPT_GET_PACK_MAX_OBJECTS
        Get the maximum number of objects to include in a pack.

    GIT_OPT_SET_PACK_MAX_OBJECTS, count
        Set the maximum number of objects to include in a pack.
    """

    result: str | None | list[str]

    if option_type in (
        C.GIT_OPT_GET_MWINDOW_SIZE,
        C.GIT_OPT_GET_MWINDOW_MAPPED_LIMIT,
        C.GIT_OPT_GET_MWINDOW_FILE_LIMIT,
    ):
        check_args(option_type, arg1, arg2, 0)

        size_ptr = ffi.new('size_t *')
        err = C.git_libgit2_opts(option_type, size_ptr)
        check_error(err)
        return size_ptr[0]

    elif option_type in (
        C.GIT_OPT_SET_MWINDOW_SIZE,
        C.GIT_OPT_SET_MWINDOW_MAPPED_LIMIT,
        C.GIT_OPT_SET_MWINDOW_FILE_LIMIT,
    ):
        check_args(option_type, arg1, arg2, 1)

        if not isinstance(arg1, int):
            raise TypeError(f'option value must be an integer, not {type(arg1)}')
        size = arg1
        if size < 0:
            raise ValueError('size must be non-negative')

        err = C.git_libgit2_opts(option_type, ffi.cast('size_t', size))
        check_error(err)
        return None

    elif option_type == C.GIT_OPT_GET_SEARCH_PATH:
        check_args(option_type, arg1, arg2, 1)

        level = int(arg1)  # Convert enum to int
        buf = ffi.new('git_buf *')
        err = C.git_libgit2_opts(option_type, ffi.cast('int', level), buf)
        check_error(err)

        try:
            if buf.ptr != ffi.NULL:
                result = to_str(ffi.string(buf.ptr))
            else:
                result = None
        finally:
            C.git_buf_dispose(buf)

        return result

    elif option_type == C.GIT_OPT_SET_SEARCH_PATH:
        check_args(option_type, arg1, arg2, 2)

        level = int(arg1)  # Convert enum to int
        path = arg2

        path_cdata: ArrayC[char] | NULL_TYPE
        if path is None:
            path_cdata = ffi.NULL
        else:
            path_bytes = to_bytes(path)
            path_cdata = ffi.new('char[]', path_bytes)

        err = C.git_libgit2_opts(option_type, ffi.cast('int', level), path_cdata)
        check_error(err)
        return None

    elif option_type == C.GIT_OPT_SET_CACHE_OBJECT_LIMIT:
        check_args(option_type, arg1, arg2, 2)

        object_type = int(arg1)  # Convert enum to int
        if not isinstance(arg2, int):
            raise TypeError(
                f'option value must be an integer, not {type(arg2).__name__}'
            )
        size = arg2
        if size < 0:
            raise ValueError('size must be non-negative')

        err = C.git_libgit2_opts(
            option_type, ffi.cast('int', object_type), ffi.cast('size_t', size)
        )
        check_error(err)
        return None

    elif option_type == C.GIT_OPT_SET_CACHE_MAX_SIZE:
        check_args(option_type, arg1, arg2, 1)

        size = arg1
        if not isinstance(size, int):
            raise TypeError(
                f'option value must be an integer, not {type(size).__name__}'
            )

        err = C.git_libgit2_opts(option_type, ffi.cast('ssize_t', size))
        check_error(err)
        return None

    elif option_type == C.GIT_OPT_GET_CACHED_MEMORY:
        check_args(option_type, arg1, arg2, 0)

        current_ptr = ffi.new('ssize_t *')
        allowed_ptr = ffi.new('ssize_t *')
        err = C.git_libgit2_opts(option_type, current_ptr, allowed_ptr)
        check_error(err)
        return (current_ptr[0], allowed_ptr[0])

    elif option_type == C.GIT_OPT_SET_SSL_CERT_LOCATIONS:
        check_args(option_type, arg1, arg2, 2)

        cert_file = arg1
        cert_dir = arg2

        cert_file_cdata: ArrayC[char] | NULL_TYPE
        if cert_file is None:
            cert_file_cdata = ffi.NULL
        else:
            cert_file_bytes = to_bytes(cert_file)
            cert_file_cdata = ffi.new('char[]', cert_file_bytes)

        cert_dir_cdata: ArrayC[char] | NULL_TYPE
        if cert_dir is None:
            cert_dir_cdata = ffi.NULL
        else:
            cert_dir_bytes = to_bytes(cert_dir)
            cert_dir_cdata = ffi.new('char[]', cert_dir_bytes)

        err = C.git_libgit2_opts(option_type, cert_file_cdata, cert_dir_cdata)
        check_error(err)
        return None

    # Handle boolean/int enable/disable options
    elif option_type in (
        C.GIT_OPT_ENABLE_CACHING,
        C.GIT_OPT_ENABLE_STRICT_OBJECT_CREATION,
        C.GIT_OPT_ENABLE_STRICT_SYMBOLIC_REF_CREATION,
        C.GIT_OPT_ENABLE_OFS_DELTA,
        C.GIT_OPT_ENABLE_FSYNC_GITDIR,
        C.GIT_OPT_ENABLE_STRICT_HASH_VERIFICATION,
        C.GIT_OPT_ENABLE_UNSAVED_INDEX_SAFETY,
        C.GIT_OPT_DISABLE_PACK_KEEP_FILE_CHECKS,
        C.GIT_OPT_SET_OWNER_VALIDATION,
    ):
        check_args(option_type, arg1, arg2, 1)

        enabled = arg1
        # Convert to int (0 or 1)
        value = 1 if enabled else 0

        err = C.git_libgit2_opts(option_type, ffi.cast('int', value))
        check_error(err)
        return None

    elif option_type == C.GIT_OPT_GET_OWNER_VALIDATION:
        check_args(option_type, arg1, arg2, 0)

        enabled_ptr = ffi.new('int *')
        err = C.git_libgit2_opts(option_type, enabled_ptr)
        check_error(err)
        return bool(enabled_ptr[0])

    elif option_type == C.GIT_OPT_GET_TEMPLATE_PATH:
        check_args(option_type, arg1, arg2, 0)

        buf = ffi.new('git_buf *')
        err = C.git_libgit2_opts(option_type, buf)
        check_error(err)

        try:
            if buf.ptr != ffi.NULL:
                result = to_str(ffi.string(buf.ptr))
            else:
                result = None
        finally:
            C.git_buf_dispose(buf)

        return result

    elif option_type == C.GIT_OPT_SET_TEMPLATE_PATH:
        check_args(option_type, arg1, arg2, 1)

        path = arg1
        template_path_cdata: ArrayC[char] | NULL_TYPE
        if path is None:
            template_path_cdata = ffi.NULL
        else:
            path_bytes = to_bytes(path)
            template_path_cdata = ffi.new('char[]', path_bytes)

        err = C.git_libgit2_opts(option_type, template_path_cdata)
        check_error(err)
        return None

    elif option_type == C.GIT_OPT_GET_USER_AGENT:
        check_args(option_type, arg1, arg2, 0)

        buf = ffi.new('git_buf *')
        err = C.git_libgit2_opts(option_type, buf)
        check_error(err)

        try:
            if buf.ptr != ffi.NULL:
                result = to_str(ffi.string(buf.ptr))
            else:
                result = None
        finally:
            C.git_buf_dispose(buf)

        return result

    elif option_type == C.GIT_OPT_SET_USER_AGENT:
        check_args(option_type, arg1, arg2, 1)

        agent = arg1
        agent_bytes = to_bytes(agent)
        agent_cdata = ffi.new('char[]', agent_bytes)

        err = C.git_libgit2_opts(option_type, agent_cdata)
        check_error(err)
        return None

    elif option_type == C.GIT_OPT_SET_SSL_CIPHERS:
        check_args(option_type, arg1, arg2, 1)

        ciphers = arg1
        ciphers_bytes = to_bytes(ciphers)
        ciphers_cdata = ffi.new('char[]', ciphers_bytes)

        err = C.git_libgit2_opts(option_type, ciphers_cdata)
        check_error(err)
        return None

    # Handle GET_WINDOWS_SHAREMODE
    elif option_type == C.GIT_OPT_GET_WINDOWS_SHAREMODE:
        check_args(option_type, arg1, arg2, 0)

        value_ptr = ffi.new('unsigned int *')
        err = C.git_libgit2_opts(option_type, value_ptr)
        check_error(err)
        return value_ptr[0]

    # Handle SET_WINDOWS_SHAREMODE
    elif option_type == C.GIT_OPT_SET_WINDOWS_SHAREMODE:
        check_args(option_type, arg1, arg2, 1)

        if not isinstance(arg1, int):
            raise TypeError(
                f'option value must be an integer, not {type(arg1).__name__}'
            )
        value = arg1
        if value < 0:
            raise ValueError('value must be non-negative')

        err = C.git_libgit2_opts(option_type, ffi.cast('unsigned int', value))
        check_error(err)
        return None

    # Handle GET_PACK_MAX_OBJECTS
    elif option_type == C.GIT_OPT_GET_PACK_MAX_OBJECTS:
        check_args(option_type, arg1, arg2, 0)

        size_ptr = ffi.new('size_t *')
        err = C.git_libgit2_opts(option_type, size_ptr)
        check_error(err)
        return size_ptr[0]

    # Handle SET_PACK_MAX_OBJECTS
    elif option_type == C.GIT_OPT_SET_PACK_MAX_OBJECTS:
        check_args(option_type, arg1, arg2, 1)

        if not isinstance(arg1, int):
            raise TypeError(
                f'option value must be an integer, not {type(arg1).__name__}'
            )
        size = arg1
        if size < 0:
            raise ValueError('size must be non-negative')

        err = C.git_libgit2_opts(option_type, ffi.cast('size_t', size))
        check_error(err)
        return None

    # Handle ENABLE_HTTP_EXPECT_CONTINUE
    elif option_type == C.GIT_OPT_ENABLE_HTTP_EXPECT_CONTINUE:
        check_args(option_type, arg1, arg2, 1)

        enabled = arg1
        # Convert to int (0 or 1)
        value = 1 if enabled else 0

        err = C.git_libgit2_opts(option_type, ffi.cast('int', value))
        check_error(err)
        return None

    # Handle SET_ODB_PACKED_PRIORITY
    elif option_type == C.GIT_OPT_SET_ODB_PACKED_PRIORITY:
        check_args(option_type, arg1, arg2, 1)

        if not isinstance(arg1, int):
            raise TypeError(
                f'option value must be an integer, not {type(arg1).__name__}'
            )
        priority = arg1

        err = C.git_libgit2_opts(option_type, ffi.cast('int', priority))
        check_error(err)
        return None

    # Handle SET_ODB_LOOSE_PRIORITY
    elif option_type == C.GIT_OPT_SET_ODB_LOOSE_PRIORITY:
        check_args(option_type, arg1, arg2, 1)

        if not isinstance(arg1, int):
            raise TypeError(
                f'option value must be an integer, not {type(arg1).__name__}'
            )
        priority = arg1

        err = C.git_libgit2_opts(option_type, ffi.cast('int', priority))
        check_error(err)
        return None

    # Handle GET_EXTENSIONS
    elif option_type == C.GIT_OPT_GET_EXTENSIONS:
        check_args(option_type, arg1, arg2, 0)

        # GET_EXTENSIONS expects a git_strarray pointer
        strarray = ffi.new('git_strarray *')
        err = C.git_libgit2_opts(option_type, strarray)
        check_error(err)

        result = []
        try:
            if strarray.strings != ffi.NULL:
                # Cast to the non-NULL type for type checking
                strings = cast('ArrayC[char_pointer]', strarray.strings)
                for i in range(strarray.count):
                    if strings[i] != ffi.NULL:
                        result.append(to_str(ffi.string(strings[i])))
        finally:
            # Must dispose of the strarray to free the memory
            C.git_strarray_dispose(strarray)

        return result

    # Handle SET_EXTENSIONS
    elif option_type == C.GIT_OPT_SET_EXTENSIONS:
        check_args(option_type, arg1, arg2, 2)

        extensions = arg1
        length = arg2

        if not isinstance(extensions, list):
            raise TypeError('extensions must be a list of strings')
        if not isinstance(length, int):
            raise TypeError('length must be an integer')

        # Create array of char pointers
        # libgit2 will make its own copies with git__strdup
        ext_array: ArrayC[char_pointer] = ffi.new('char *[]', len(extensions))
        ext_strings: list[ArrayC[char]] = []  # Keep references during the call

        for i, ext in enumerate(extensions):
            ext_bytes = to_bytes(ext)
            ext_string: ArrayC[char] = ffi.new('char[]', ext_bytes)
            ext_strings.append(ext_string)
            ext_array[i] = ffi.cast('char *', ext_string)

        err = C.git_libgit2_opts(option_type, ext_array, ffi.cast('size_t', length))
        check_error(err)
        return None

    # Handle GET_HOMEDIR
    elif option_type == C.GIT_OPT_GET_HOMEDIR:
        check_args(option_type, arg1, arg2, 0)

        buf = ffi.new('git_buf *')
        err = C.git_libgit2_opts(option_type, buf)
        check_error(err)

        try:
            if buf.ptr != ffi.NULL:
                result = to_str(ffi.string(buf.ptr))
            else:
                result = None
        finally:
            C.git_buf_dispose(buf)

        return result

    # Handle SET_HOMEDIR
    elif option_type == C.GIT_OPT_SET_HOMEDIR:
        check_args(option_type, arg1, arg2, 1)

        path = arg1
        homedir_cdata: ArrayC[char] | NULL_TYPE
        if path is None:
            homedir_cdata = ffi.NULL
        else:
            path_bytes = to_bytes(path)
            homedir_cdata = ffi.new('char[]', path_bytes)

        err = C.git_libgit2_opts(option_type, homedir_cdata)
        check_error(err)
        return None

    # Handle GET_SERVER_CONNECT_TIMEOUT
    elif option_type == C.GIT_OPT_GET_SERVER_CONNECT_TIMEOUT:
        check_args(option_type, arg1, arg2, 0)

        timeout_ptr = ffi.new('int *')
        err = C.git_libgit2_opts(option_type, timeout_ptr)
        check_error(err)
        return timeout_ptr[0]

    # Handle SET_SERVER_CONNECT_TIMEOUT
    elif option_type == C.GIT_OPT_SET_SERVER_CONNECT_TIMEOUT:
        check_args(option_type, arg1, arg2, 1)

        if not isinstance(arg1, int):
            raise TypeError(
                f'option value must be an integer, not {type(arg1).__name__}'
            )
        timeout = arg1

        err = C.git_libgit2_opts(option_type, ffi.cast('int', timeout))
        check_error(err)
        return None

    # Handle GET_SERVER_TIMEOUT
    elif option_type == C.GIT_OPT_GET_SERVER_TIMEOUT:
        check_args(option_type, arg1, arg2, 0)

        timeout_ptr = ffi.new('int *')
        err = C.git_libgit2_opts(option_type, timeout_ptr)
        check_error(err)
        return timeout_ptr[0]

    # Handle SET_SERVER_TIMEOUT
    elif option_type == C.GIT_OPT_SET_SERVER_TIMEOUT:
        check_args(option_type, arg1, arg2, 1)

        if not isinstance(arg1, int):
            raise TypeError(
                f'option value must be an integer, not {type(arg1).__name__}'
            )
        timeout = arg1

        err = C.git_libgit2_opts(option_type, ffi.cast('int', timeout))
        check_error(err)
        return None

    # Handle GET_USER_AGENT_PRODUCT
    elif option_type == C.GIT_OPT_GET_USER_AGENT_PRODUCT:
        check_args(option_type, arg1, arg2, 0)

        buf = ffi.new('git_buf *')
        err = C.git_libgit2_opts(option_type, buf)
        check_error(err)

        try:
            if buf.ptr != ffi.NULL:
                result = to_str(ffi.string(buf.ptr))
            else:
                result = None
        finally:
            C.git_buf_dispose(buf)

        return result

    # Handle SET_USER_AGENT_PRODUCT
    elif option_type == C.GIT_OPT_SET_USER_AGENT_PRODUCT:
        check_args(option_type, arg1, arg2, 1)

        product = arg1
        product_bytes = to_bytes(product)
        product_cdata = ffi.new('char[]', product_bytes)

        err = C.git_libgit2_opts(option_type, product_cdata)
        check_error(err)
        return None

    # Not implemented - ADD_SSL_X509_CERT requires directly binding with OpenSSL
    # as the API works accepts a X509* struct.  Use GIT_OPT_SET_SSL_CERT_LOCATIONS
    # instead.
    elif option_type == C.GIT_OPT_ADD_SSL_X509_CERT:
        raise NotImplementedError('Use GIT_OPT_SET_SSL_CERT_LOCATIONS instead')

    # Not implemented - SET_ALLOCATOR is not feasible from Python level
    # because it requires providing C function pointers for memory management
    # (malloc, free, etc.) that must handle raw memory at the C level,
    # which cannot be safely implemented in pure Python.
    elif option_type == C.GIT_OPT_SET_ALLOCATOR:
        raise NotImplementedError('Setting a custom allocator not possible from Python')

    else:
        raise ValueError(f'Invalid option {option_type}')



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import typing
import warnings
from dataclasses import dataclass
from os import PathLike

# Import from pygit2
from ._pygit2 import Diff, Oid, Tree
from .enums import DiffOption, FileMode
from .errors import check_error
from .ffi import C, ffi
from .utils import GenericIterator, StrArray, to_bytes, to_str

if typing.TYPE_CHECKING:
    from .repository import Repository


class Index:
    # XXX Implement the basic features in C (_pygit2.Index) and make
    # pygit2.Index to inherit from _pygit2.Index? This would allow for
    # a proper implementation in some places: e.g. checking the index type
    # from C code (see Tree_diff_to_index)

    def __init__(self, path: str | PathLike[str] | None = None) -> None:
        """Create a new Index

        If path is supplied, the read and write methods will use that path
        to read from and write to.
        """
        cindex = ffi.new('git_index **')
        err = C.git_index_open(cindex, to_bytes(path))
        check_error(err)

        self._repo = None
        self._index = cindex[0]
        self._cindex = cindex

    @classmethod
    def from_c(cls, repo, ptr):
        index = cls.__new__(cls)
        index._repo = repo
        index._index = ptr[0]
        index._cindex = ptr

        return index

    @property
    def _pointer(self):
        return bytes(ffi.buffer(self._cindex)[:])

    def __del__(self) -> None:
        C.git_index_free(self._index)

    def __len__(self) -> int:
        return C.git_index_entrycount(self._index)

    def __contains__(self, path) -> bool:
        err = C.git_index_find(ffi.NULL, self._index, to_bytes(path))
        if err == C.GIT_ENOTFOUND:
            return False

        check_error(err)
        return True

    def __getitem__(self, key: str | int | PathLike[str]) -> 'IndexEntry':
        centry = ffi.NULL
        if isinstance(key, str) or hasattr(key, '__fspath__'):
            centry = C.git_index_get_bypath(self._index, to_bytes(key), 0)
        elif isinstance(key, int):
            if key >= 0:
                centry = C.git_index_get_byindex(self._index, key)
            else:
                raise ValueError(key)
        else:
            raise TypeError(f'Expected str or int, got {type(key)}')

        if centry == ffi.NULL:
            raise KeyError(key)

        return IndexEntry._from_c(centry)

    def __iter__(self):
        return GenericIterator(self)

    def read(self, force: bool = True) -> None:
        """
        Update the contents of the Index by reading from a file.

        Parameters:

        force
            If True (the default) always reload. If False, only if the file
            has changed.
        """

        err = C.git_index_read(self._index, force)
        check_error(err, io=True)

    def write(self) -> None:
        """Write the contents of the Index to disk."""
        err = C.git_index_write(self._index)
        check_error(err, io=True)

    def clear(self) -> None:
        err = C.git_index_clear(self._index)
        check_error(err)

    def read_tree(self, tree: Oid | Tree | str) -> None:
        """Replace the contents of the Index with those of the given tree,
        expressed either as a <Tree> object or as an oid (string or <Oid>).

        The tree will be read recursively and all its children will also be
        inserted into the Index.
        """
        repo = self._repo
        if isinstance(tree, str):
            if repo is None:
                raise TypeError('id given but no associated repository')
            tree = repo[tree]

        if isinstance(tree, Oid):
            if repo is None:
                raise TypeError('id given but no associated repository')

            tree = repo[tree]
        elif not isinstance(tree, Tree):
            raise TypeError('argument must be Oid, Tree or str')

        tree_cptr = ffi.new('git_tree **')
        ffi.buffer(tree_cptr)[:] = tree._pointer[:]
        err = C.git_index_read_tree(self._index, tree_cptr[0])
        check_error(err)

    def write_tree(self, repo: 'Repository | None' = None) -> Oid:
        """Create a tree out of the Index. Return the <Oid> object of the
        written tree.

        The contents of the index will be written out to the object
        database. If there is no associated repository, 'repo' must be
        passed. If there is an associated repository and 'repo' is
        passed, then that repository will be used instead.

        It returns the id of the resulting tree.
        """
        coid = ffi.new('git_oid *')

        repo = repo or self._repo

        if repo:
            err = C.git_index_write_tree_to(coid, self._index, repo._repo)
        else:
            err = C.git_index_write_tree(coid, self._index)

        check_error(err)
        return Oid(raw=bytes(ffi.buffer(coid)[:]))

    def remove(self, path: PathLike[str] | str, level: int = 0) -> None:
        """Remove an entry from the Index."""
        err = C.git_index_remove(self._index, to_bytes(path), level)
        check_error(err, io=True)

    def remove_directory(self, path: PathLike[str] | str, level: int = 0) -> None:
        """Remove a directory from the Index."""
        err = C.git_index_remove_directory(self._index, to_bytes(path), level)
        check_error(err, io=True)

    def remove_all(self, pathspecs: typing.Sequence[str | PathLike[str]]) -> None:
        """Remove all index entries matching pathspecs."""
        with StrArray(pathspecs) as arr:
            err = C.git_index_remove_all(self._index, arr.ptr, ffi.NULL, ffi.NULL)
            check_error(err, io=True)

    def add_all(self, pathspecs: None | list[str | PathLike[str]] = None) -> None:
        """Add or update index entries matching files in the working directory.

        If pathspecs are specified, only files matching those pathspecs will
        be added.
        """
        pathspecs = pathspecs or []
        with StrArray(pathspecs) as arr:
            err = C.git_index_add_all(self._index, arr.ptr, 0, ffi.NULL, ffi.NULL)
            check_error(err, io=True)

    def add(self, path_or_entry: 'IndexEntry | str | PathLike[str]') -> None:
        """Add or update an entry in the Index.

        If a path is given, that file will be added. The path must be relative
        to the root of the worktree and the Index must be associated with a
        repository.

        If an IndexEntry is given, that entry will be added or update in the
        Index without checking for the existence of the path or id.
        """
        if isinstance(path_or_entry, IndexEntry):
            entry = path_or_entry
            centry, str_ref = entry._to_c()
            err = C.git_index_add(self._index, centry)
        elif isinstance(path_or_entry, str) or hasattr(path_or_entry, '__fspath__'):
            path = path_or_entry
            err = C.git_index_add_bypath(self._index, to_bytes(path))
        else:
            raise TypeError('argument must be string, Path or IndexEntry')

        check_error(err, io=True)

    def add_conflict(
        self, ancestor: 'IndexEntry', ours: 'IndexEntry', theirs: 'IndexEntry | None'
    ) -> None:
        """
        Add or update index entries to represent a conflict. Any staged entries that
        exist at the given paths will be removed.

        Parameters:

        ancestor
            ancestor of the conflict
        ours
            ours side of the conflict
        theirs
            their side of the conflict
        """

        if ancestor and not isinstance(ancestor, IndexEntry):
            raise TypeError('ancestor has to be an instance of IndexEntry or None')
        if ours and not isinstance(ours, IndexEntry):
            raise TypeError('ours has to be an instance of IndexEntry or None')
        if theirs and not isinstance(theirs, IndexEntry):
            raise TypeError('theirs has to be an instance of IndexEntry or None')

        centry_ancestor: ffi.NULL_TYPE | ffi.GitIndexEntryC = ffi.NULL
        centry_ours: ffi.NULL_TYPE | ffi.GitIndexEntryC = ffi.NULL
        centry_theirs: ffi.NULL_TYPE | ffi.GitIndexEntryC = ffi.NULL
        if ancestor is not None:
            centry_ancestor, _ = ancestor._to_c()
        if ours is not None:
            centry_ours, _ = ours._to_c()
        if theirs is not None:
            centry_theirs, _ = theirs._to_c()
        err = C.git_index_conflict_add(
            self._index, centry_ancestor, centry_ours, centry_theirs
        )

        check_error(err, io=True)

    def diff_to_workdir(
        self,
        flags: DiffOption = DiffOption.NORMAL,
        context_lines: int = 3,
        interhunk_lines: int = 0,
    ) -> Diff:
        """
        Diff the index against the working directory. Return a <Diff> object
        with the differences between the index and the working copy.

        Parameters:

        flags
            A combination of enums.DiffOption constants.

        context_lines
            The number of unchanged lines that define the boundary of a hunk
            (and to display before and after).

        interhunk_lines
            The maximum number of unchanged lines between hunk boundaries
            before the hunks will be merged into a one.
        """
        repo = self._repo
        if repo is None:
            raise ValueError('diff needs an associated repository')

        copts = ffi.new('git_diff_options *')
        err = C.git_diff_options_init(copts, 1)
        check_error(err)

        copts.flags = int(flags)
        copts.context_lines = context_lines
        copts.interhunk_lines = interhunk_lines

        cdiff = ffi.new('git_diff **')
        err = C.git_diff_index_to_workdir(cdiff, repo._repo, self._index, copts)
        check_error(err)

        return Diff.from_c(bytes(ffi.buffer(cdiff)[:]), repo)

    def diff_to_tree(
        self,
        tree: Tree,
        flags: DiffOption = DiffOption.NORMAL,
        context_lines: int = 3,
        interhunk_lines: int = 0,
    ) -> Diff:
        """
        Diff the index against a tree.  Return a <Diff> object with the
        differences between the index and the given tree.

        Parameters:

        tree
            The tree to diff.

        flags
            A combination of enums.DiffOption constants.

        context_lines
            The number of unchanged lines that define the boundary of a hunk
            (and to display before and after).

        interhunk_lines
            The maximum number of unchanged lines between hunk boundaries
            before the hunks will be merged into a one.
        """
        repo = self._repo
        if repo is None:
            raise ValueError('diff needs an associated repository')

        if not isinstance(tree, Tree):
            raise TypeError('tree must be a Tree')

        copts = ffi.new('git_diff_options *')
        err = C.git_diff_options_init(copts, 1)
        check_error(err)

        copts.flags = int(flags)
        copts.context_lines = context_lines
        copts.interhunk_lines = interhunk_lines

        ctree = ffi.new('git_tree **')
        ffi.buffer(ctree)[:] = tree._pointer[:]

        cdiff = ffi.new('git_diff **')
        err = C.git_diff_tree_to_index(cdiff, repo._repo, ctree[0], self._index, copts)
        check_error(err)

        return Diff.from_c(bytes(ffi.buffer(cdiff)[:]), repo)

    #
    # Conflicts
    #

    @property
    def conflicts(self):
        """A collection of conflict information

        If there are no conflicts None is returned. Otherwise return an object
        that represents the conflicts in the index.

        This object presents a mapping interface with the paths as keys. You
        can use the ``del`` operator to remove a conflict from the Index.

        Each conflict is made up of three elements. Access or iteration
        of the conflicts returns a three-tuple of
        :py:class:`~pygit2.IndexEntry`. The first is the common
        ancestor, the second is the "ours" side of the conflict, and the
        third is the "theirs" side.

        These elements may be None depending on which sides exist for
        the particular conflict.
        """
        if not C.git_index_has_conflicts(self._index):
            return None

        return ConflictCollection(self)


@dataclass
class MergeFileResult:
    automergeable: bool
    'True if the output was automerged, false if the output contains conflict markers'

    path: str | None | PathLike[str]
    'The path that the resultant merge file should use, or None if a filename conflict would occur'

    mode: FileMode
    'The mode that the resultant merge file should use'

    contents: str
    'Contents of the file, which might include conflict markers'

    def __repr__(self):
        t = type(self)
        contents = (
            self.contents if len(self.contents) <= 20 else f'{self.contents[:20]}...'
        )
        return (
            f'<{t.__module__}.{t.__qualname__} "'
            f'automergeable={self.automergeable} "'
            f'path={self.path} '
            f'mode={self.mode} '
            f'contents={contents}>'
        )

    @classmethod
    def _from_c(cls, centry):
        if centry == ffi.NULL:
            return None

        automergeable = centry.automergeable != 0
        path = to_str(ffi.string(centry.path)) if centry.path else None
        mode = FileMode(centry.mode)
        contents = ffi.string(centry.ptr, centry.len).decode('utf-8')

        return MergeFileResult(automergeable, path, mode, contents)


class IndexEntry:
    path: str | PathLike[str]
    'The path of this entry'

    id: Oid
    'The id of the referenced object'

    mode: FileMode
    'The mode of this entry, a FileMode value'

    def __init__(
        self, path: str | PathLike[str], object_id: Oid, mode: FileMode
    ) -> None:
        self.path = path
        self.id = object_id
        self.mode = mode

    @property
    def oid(self):
        warnings.warn('Use entry.id', DeprecationWarning)
        return self.id

    def __str__(self):
        return f'<path={self.path} id={self.id} mode={self.mode}>'

    def __repr__(self):
        t = type(self)
        return f'<{t.__module__}.{t.__qualname__} path={self.path} id={self.id} mode={self.mode}>'

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, IndexEntry):
            return NotImplemented
        return (
            self.path == other.path and self.id == other.id and self.mode == other.mode
        )

    def _to_c(self) -> tuple['ffi.GitIndexEntryC', 'ffi.ArrayC[ffi.char]']:
        """Convert this entry into the C structure

        The first returned arg is the pointer, the second is the reference to
        the string we allocated, which we need to exist past this function
        """
        centry = ffi.new('git_index_entry *')
        # basically memcpy()
        ffi.buffer(ffi.addressof(centry, 'id'))[:] = self.id.raw[:]
        centry.mode = int(self.mode)
        path = ffi.new('char[]', to_bytes(self.path))
        centry.path = path

        return centry, path

    @classmethod
    def _from_c(cls, centry):
        if centry == ffi.NULL:
            return None

        entry = cls.__new__(cls)
        entry.path = to_str(ffi.string(centry.path))
        entry.mode = FileMode(centry.mode)
        entry.id = Oid(raw=bytes(ffi.buffer(ffi.addressof(centry, 'id'))[:]))

        return entry


class ConflictCollection:
    def __init__(self, index):
        self._index = index

    def __getitem__(self, path):
        cancestor = ffi.new('git_index_entry **')
        cours = ffi.new('git_index_entry **')
        ctheirs = ffi.new('git_index_entry **')

        err = C.git_index_conflict_get(
            cancestor, cours, ctheirs, self._index._index, to_bytes(path)
        )
        check_error(err)

        ancestor = IndexEntry._from_c(cancestor[0])
        ours = IndexEntry._from_c(cours[0])
        theirs = IndexEntry._from_c(ctheirs[0])

        return ancestor, ours, theirs

    def __delitem__(self, path):
        err = C.git_index_conflict_remove(self._index._index, to_bytes(path))
        check_error(err)

    def __iter__(self):
        return ConflictIterator(self._index)

    def __contains__(self, path):
        cancestor = ffi.new('git_index_entry **')
        cours = ffi.new('git_index_entry **')
        ctheirs = ffi.new('git_index_entry **')

        err = C.git_index_conflict_get(
            cancestor, cours, ctheirs, self._index._index, to_bytes(path)
        )
        if err == C.GIT_ENOTFOUND:
            return False

        check_error(err)
        return True


class ConflictIterator:
    def __init__(self, index):
        citer = ffi.new('git_index_conflict_iterator **')
        err = C.git_index_conflict_iterator_new(citer, index._index)
        check_error(err)
        self._index = index
        self._iter = citer[0]

    def __del__(self):
        C.git_index_conflict_iterator_free(self._iter)

    def __iter__(self):
        return self

    def __next__(self):
        cancestor = ffi.new('git_index_entry **')
        cours = ffi.new('git_index_entry **')
        ctheirs = ffi.new('git_index_entry **')

        err = C.git_index_conflict_next(cancestor, cours, ctheirs, self._iter)
        if err == C.GIT_ITEROVER:
            raise StopIteration

        check_error(err)

        ancestor = IndexEntry._from_c(cancestor[0])
        ours = IndexEntry._from_c(cours[0])
        theirs = IndexEntry._from_c(ctheirs[0])

        return ancestor, ours, theirs



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from __future__ import annotations

import threading
from typing import TYPE_CHECKING

from .errors import check_error
from .ffi import C, ffi
from .utils import to_bytes

if TYPE_CHECKING:
    from ._pygit2 import Oid, Signature
    from .repository import BaseRepository


class ReferenceTransaction:
    """Context manager for transactional reference updates.

    A transaction allows multiple reference updates to be performed atomically.
    All updates are applied when the transaction is committed, or none are applied
    if the transaction is rolled back.

    Example:
        with repo.transaction() as txn:
            txn.lock_ref('refs/heads/master')
            txn.set_target('refs/heads/master', new_oid, message='Update master')
            # Changes committed automatically on context exit
    """

    def __init__(self, repository: BaseRepository) -> None:
        self._repository = repository
        self._transaction = ffi.new('git_transaction **')
        self._tx = None
        self._thread_id = threading.get_ident()

        err = C.git_transaction_new(self._transaction, repository._repo)
        check_error(err)
        self._tx = self._transaction[0]

    def _check_thread(self) -> None:
        """Verify transaction is being used from the same thread that created it."""
        current_thread = threading.get_ident()
        if current_thread != self._thread_id:
            raise RuntimeError(
                f'Transaction created in thread {self._thread_id} '
                f'but used in thread {current_thread}. '
                'Transactions must be used from the thread that created them.'
            )

    def lock_ref(self, refname: str) -> None:
        """Lock a reference in preparation for updating it.

        Args:
            refname: Name of the reference to lock (e.g., 'refs/heads/master')
        """
        self._check_thread()
        if self._tx is None:
            raise ValueError('Transaction already closed')

        c_refname = ffi.new('char[]', to_bytes(refname))
        err = C.git_transaction_lock_ref(self._tx, c_refname)
        check_error(err)

    def set_target(
        self,
        refname: str,
        target: Oid | str,
        signature: Signature | None = None,
        message: str | None = None,
    ) -> None:
        """Set the target of a direct reference.

        The reference must be locked first via lock_ref().

        Args:
            refname: Name of the reference to update
            target: Target OID or hex string
            signature: Signature for the reflog (None to use repo identity)
            message: Message for the reflog
        """
        self._check_thread()
        if self._tx is None:
            raise ValueError('Transaction already closed')

        from ._pygit2 import Oid

        c_refname = ffi.new('char[]', to_bytes(refname))

        # Convert target to OID
        if isinstance(target, str):
            target = Oid(hex=target)

        c_oid = ffi.new('git_oid *')
        ffi.buffer(c_oid)[:] = target.raw

        c_sig = signature._pointer if signature else ffi.NULL
        c_msg = ffi.new('char[]', to_bytes(message)) if message else ffi.NULL

        err = C.git_transaction_set_target(self._tx, c_refname, c_oid, c_sig, c_msg)
        check_error(err)

    def set_symbolic_target(
        self,
        refname: str,
        target: str,
        signature: Signature | None = None,
        message: str | None = None,
    ) -> None:
        """Set the target of a symbolic reference.

        The reference must be locked first via lock_ref().

        Args:
            refname: Name of the reference to update
            target: Target reference name (e.g., 'refs/heads/master')
            signature: Signature for the reflog (None to use repo identity)
            message: Message for the reflog
        """
        self._check_thread()
        if self._tx is None:
            raise ValueError('Transaction already closed')

        c_refname = ffi.new('char[]', to_bytes(refname))
        c_target = ffi.new('char[]', to_bytes(target))
        c_sig = signature._pointer if signature else ffi.NULL
        c_msg = ffi.new('char[]', to_bytes(message)) if message else ffi.NULL

        err = C.git_transaction_set_symbolic_target(
            self._tx, c_refname, c_target, c_sig, c_msg
        )
        check_error(err)

    def remove(self, refname: str) -> None:
        """Remove a reference.

        The reference must be locked first via lock_ref().

        Args:
            refname: Name of the reference to remove
        """
        self._check_thread()
        if self._tx is None:
            raise ValueError('Transaction already closed')

        c_refname = ffi.new('char[]', to_bytes(refname))
        err = C.git_transaction_remove(self._tx, c_refname)
        check_error(err)

    def commit(self) -> None:
        """Commit the transaction, applying all queued updates."""
        self._check_thread()
        if self._tx is None:
            raise ValueError('Transaction already closed')

        err = C.git_transaction_commit(self._tx)
        check_error(err)

    def __enter__(self) -> ReferenceTransaction:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._check_thread()
        # Only commit if no exception occurred
        if exc_type is None and self._tx is not None:
            self.commit()

        # Always free the transaction
        if self._tx is not None:
            C.git_transaction_free(self._tx)
            self._tx = None

    def __del__(self) -> None:
        if self._tx is not None:
            C.git_transaction_free(self._tx)
            self._tx = None



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from collections.abc import Callable, Iterator
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property  # type: ignore

# Import from pygit2
from .errors import check_error
from .ffi import C, ffi
from .utils import to_bytes

if TYPE_CHECKING:
    from ._libgit2.ffi import GitConfigC, GitConfigEntryC
    from .repository import BaseRepository


def str_to_bytes(value: str | PathLike[str] | bytes, name: str) -> bytes:
    if not isinstance(value, str):
        raise TypeError(f'{name} must be a string')

    return to_bytes(value)


class ConfigIterator:
    def __init__(self, config, ptr) -> None:
        self._iter = ptr
        self._config = config

    def __del__(self) -> None:
        C.git_config_iterator_free(self._iter)

    def __iter__(self) -> 'ConfigIterator':
        return self

    def __next__(self) -> 'ConfigEntry':
        return self._next_entry()

    def _next_entry(self) -> 'ConfigEntry':
        centry = ffi.new('git_config_entry **')
        err = C.git_config_next(centry, self._iter)
        check_error(err)

        return ConfigEntry._from_c(centry[0], self)


class ConfigMultivarIterator(ConfigIterator):
    def __next__(self) -> str | None:  # type: ignore[override]
        entry = self._next_entry()
        return entry.value


class Config:
    """Git configuration management."""

    _repo: 'BaseRepository'
    _config: 'GitConfigC'

    def __init__(self, path: str | None = None) -> None:
        cconfig = ffi.new('git_config **')

        if not path:
            err = C.git_config_new(cconfig)
        else:
            path_bytes = str_to_bytes(path, 'path')
            err = C.git_config_open_ondisk(cconfig, path_bytes)

        check_error(err, io=True)
        self._config = cconfig[0]

    @classmethod
    def from_c(cls, repo: 'BaseRepository', ptr: 'GitConfigC') -> 'Config':
        config = cls.__new__(cls)
        config._repo = repo
        config._config = ptr

        return config

    def __del__(self) -> None:
        try:
            C.git_config_free(self._config)
        except AttributeError:
            pass

    def _get(self, key: str | bytes) -> tuple[int, 'ConfigEntry']:
        key = str_to_bytes(key, 'key')

        entry = ffi.new('git_config_entry **')
        err = C.git_config_get_entry(entry, self._config, key)

        return err, ConfigEntry._from_c(entry[0])

    def _get_entry(self, key: str | bytes) -> 'ConfigEntry':
        err, entry = self._get(key)

        if err == C.GIT_ENOTFOUND:
            raise KeyError(key)

        check_error(err)
        return entry

    def __contains__(self, key: str | bytes) -> bool:
        err, cstr = self._get(key)

        if err == C.GIT_ENOTFOUND:
            return False

        check_error(err)

        return True

    def __getitem__(self, key: str | bytes) -> str | None:
        """
        When using the mapping interface, the value is returned as a string. In
        order to apply the git-config parsing rules, you can use
        :meth:`Config.get_bool` or :meth:`Config.get_int`.
        """
        entry = self._get_entry(key)

        return entry.value

    def __setitem__(self, key: str | bytes, value: bool | int | str | bytes) -> None:
        key = str_to_bytes(key, 'key')

        err = 0
        if isinstance(value, bool):
            err = C.git_config_set_bool(self._config, key, value)
        elif isinstance(value, int):
            err = C.git_config_set_int64(self._config, key, value)
        else:
            err = C.git_config_set_string(self._config, key, to_bytes(value))

        check_error(err)

    def __delitem__(self, key: str | bytes) -> None:
        key = str_to_bytes(key, 'key')

        err = C.git_config_delete_entry(self._config, key)
        check_error(err)

    def __iter__(self) -> Iterator['ConfigEntry']:
        """
        Iterate over configuration entries, returning a ``ConfigEntry``
        objects. These contain the name, level, and value of each configuration
        variable. Be aware that this may return multiple versions of each entry
        if they are set multiple times in the configuration files.
        """
        citer = ffi.new('git_config_iterator **')
        err = C.git_config_iterator_new(citer, self._config)
        check_error(err)

        return ConfigIterator(self, citer[0])

    def get_multivar(
        self, name: str | bytes, regex: str | None = None
    ) -> ConfigMultivarIterator:
        """Get each value of a multivar ''name'' as a list of strings.

        The optional ''regex'' parameter is expected to be a regular expression
        to filter the variables we're interested in.
        """
        name = str_to_bytes(name, 'name')
        regex_bytes = to_bytes(regex or None)

        citer = ffi.new('git_config_iterator **')
        err = C.git_config_multivar_iterator_new(citer, self._config, name, regex_bytes)
        check_error(err)

        return ConfigMultivarIterator(self, citer[0])

    def set_multivar(
        self, name: str | bytes, regex: str | bytes, value: str | bytes
    ) -> None:
        """Set a multivar ''name'' to ''value''. ''regexp'' is a regular
        expression to indicate which values to replace.
        """
        name = str_to_bytes(name, 'name')
        regex = str_to_bytes(regex, 'regex')
        value = str_to_bytes(value, 'value')

        err = C.git_config_set_multivar(self._config, name, regex, value)
        check_error(err)

    def delete_multivar(self, name: str | bytes, regex: str | bytes) -> None:
        """Delete a multivar ''name''. ''regexp'' is a regular expression to
        indicate which values to delete.
        """
        name = str_to_bytes(name, 'name')
        regex = str_to_bytes(regex, 'regex')

        err = C.git_config_delete_multivar(self._config, name, regex)
        check_error(err)

    def get_bool(self, key: str | bytes) -> bool:
        """Look up *key* and parse its value as a boolean as per the git-config
        rules. Return a boolean value (True or False).

        Truthy values are: 'true', 1, 'on' or 'yes'. Falsy values are: 'false',
        0, 'off' and 'no'
        """

        entry = self._get_entry(key)
        res = ffi.new('int *')
        err = C.git_config_parse_bool(res, entry.c_value)
        check_error(err)

        return res[0] != 0

    def get_int(self, key: bytes | str) -> int:
        """Look up *key* and parse its value as an integer as per the git-config
        rules. Return an integer.

        A value can have a suffix 'k', 'm' or 'g' which stand for 'kilo',
        'mega' and 'giga' respectively.
        """

        entry = self._get_entry(key)
        res = ffi.new('int64_t *')
        err = C.git_config_parse_int64(res, entry.c_value)
        check_error(err)

        return res[0]

    def add_file(self, path: str | Path, level: int = 0, force: int = 0) -> None:
        """Add a config file instance to an existing config."""

        err = C.git_config_add_file_ondisk(
            self._config, to_bytes(path), level, ffi.NULL, force
        )
        check_error(err)

    def snapshot(self) -> 'Config':
        """Create a snapshot from this Config object.

        This means that looking up multiple values will use the same version
        of the configuration files.
        """
        ccfg = ffi.new('git_config **')
        err = C.git_config_snapshot(ccfg, self._config)
        check_error(err)

        return Config.from_c(self._repo, ccfg[0])

    #
    # Methods to parse a string according to the git-config rules
    #

    @staticmethod
    def parse_bool(text: str) -> bool:
        res = ffi.new('int *')
        err = C.git_config_parse_bool(res, to_bytes(text))
        check_error(err)

        return res[0] != 0

    @staticmethod
    def parse_int(text: str) -> int:
        res = ffi.new('int64_t *')
        err = C.git_config_parse_int64(res, to_bytes(text))
        check_error(err)

        return res[0]

    #
    # Static methods to get specialized version of the config
    #

    @staticmethod
    def _from_found_config(fn: Callable) -> 'Config':
        buf = ffi.new('git_buf *', (ffi.NULL, 0))
        err = fn(buf)
        check_error(err, io=True)
        cpath = ffi.string(buf.ptr).decode('utf-8')
        C.git_buf_dispose(buf)

        return Config(cpath)

    @staticmethod
    def get_system_config() -> 'Config':
        """Return a <Config> object representing the system configuration file."""
        return Config._from_found_config(C.git_config_find_system)

    @staticmethod
    def get_global_config() -> 'Config':
        """Return a <Config> object representing the global configuration file."""
        return Config._from_found_config(C.git_config_find_global)

    @staticmethod
    def get_xdg_config() -> 'Config':
        """Return a <Config> object representing the global configuration file."""
        return Config._from_found_config(C.git_config_find_xdg)


class ConfigEntry:
    """An entry in a configuration object."""

    _entry: 'GitConfigEntryC'
    iterator: ConfigIterator | None

    @classmethod
    def _from_c(
        cls, ptr: 'GitConfigEntryC', iterator: ConfigIterator | None = None
    ) -> 'ConfigEntry':
        """Builds the entry from a ``git_config_entry`` pointer.

        ``iterator`` must be a ``ConfigIterator`` instance if the entry was
        created during ``git_config_iterator`` actions.
        """
        entry = cls.__new__(cls)
        entry._entry = ptr
        entry.iterator = iterator

        # It should be enough to keep a reference to iterator, so we only call
        # git_config_iterator_free when we've deleted all ConfigEntry objects.
        # But it's not, to reproduce the error comment the lines below and run
        # the script in https://github.com/libgit2/pygit2/issues/970
        # So instead we load the Python object immediately. Ideally we should
        # investigate libgit2 source code.
        if iterator is not None:
            entry.raw_name = entry.raw_name
            entry.raw_value = entry.raw_value
            entry.level = entry.level

        return entry

    def __del__(self) -> None:
        if self.iterator is None:
            C.git_config_entry_free(self._entry)

    @property
    def c_value(self) -> 'ffi.char_pointer':
        """The raw ``cData`` entry value."""
        return self._entry.value

    @cached_property
    def raw_name(self) -> bytes:
        return ffi.string(self._entry.name)

    @cached_property
    def raw_value(self) -> bytes | None:
        return ffi.string(self.c_value) if self.c_value != ffi.NULL else None

    @cached_property
    def level(self) -> int:
        """The entry's ``git_config_level_t`` value."""
        return self._entry.level

    @property
    def name(self) -> str:
        """The entry's name."""
        return self.raw_name.decode('utf-8')

    @property
    def value(self) -> str | None:
        """The entry's value as a string."""
        return self.raw_value.decode('utf-8') if self.raw_value is not None else None



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""
Settings mapping.
"""

from ssl import DefaultVerifyPaths, get_default_verify_paths
from typing import overload

import pygit2.enums

from .enums import ConfigLevel, Option
from .errors import GitError
from .options import option


class SearchPathList:
    def __getitem__(self, key: ConfigLevel) -> str:
        return option(Option.GET_SEARCH_PATH, key)

    def __setitem__(self, key: ConfigLevel, value: str) -> None:
        option(Option.SET_SEARCH_PATH, key, value)


class Settings:
    """Library-wide settings interface."""

    __slots__ = '_default_tls_verify_paths', '_ssl_cert_dir', '_ssl_cert_file'

    _search_path = SearchPathList()
    _default_tls_verify_paths: DefaultVerifyPaths | None
    _ssl_cert_file: str | bytes | None
    _ssl_cert_dir: str | bytes | None

    def __init__(self) -> None:
        """Initialize global pygit2 and libgit2 settings."""
        self._initialize_tls_certificate_locations()

    def _initialize_tls_certificate_locations(self) -> None:
        """Set up initial TLS file and directory lookup locations."""
        self._default_tls_verify_paths = get_default_verify_paths()
        try:
            self.set_ssl_cert_locations(
                self._default_tls_verify_paths.cafile,
                self._default_tls_verify_paths.capath,
            )
        except GitError as git_err:
            valid_msg = "TLS backend doesn't support certificate locations"
            if str(git_err) != valid_msg:
                raise
            self._default_tls_verify_paths = None
            self._ssl_cert_file = None
            self._ssl_cert_dir = None

    @property
    def search_path(self) -> SearchPathList:
        """Configuration file search path.

        This behaves like an array whose indices correspond to ConfigLevel values.
        The local search path cannot be changed.
        """
        return self._search_path

    @property
    def mwindow_size(self) -> int:
        """Get or set the maximum mmap window size"""
        return option(Option.GET_MWINDOW_SIZE)

    @mwindow_size.setter
    def mwindow_size(self, value: int) -> None:
        option(Option.SET_MWINDOW_SIZE, value)

    @property
    def mwindow_mapped_limit(self) -> int:
        """
        Get or set the maximum memory that will be mapped in total by the
        library
        """
        return option(Option.GET_MWINDOW_MAPPED_LIMIT)

    @mwindow_mapped_limit.setter
    def mwindow_mapped_limit(self, value: int) -> None:
        option(Option.SET_MWINDOW_MAPPED_LIMIT, value)

    @property
    def mwindow_file_limit(self) -> int:
        """Get or set the maximum number of files to be mapped at any time"""
        return option(Option.GET_MWINDOW_FILE_LIMIT)

    @mwindow_file_limit.setter
    def mwindow_file_limit(self, value: int) -> None:
        option(Option.SET_MWINDOW_FILE_LIMIT, value)

    @property
    def cached_memory(self) -> tuple[int, int]:
        """
        Get the current bytes in cache and the maximum that would be
        allowed in the cache.
        """
        return option(Option.GET_CACHED_MEMORY)

    def enable_caching(self, value: bool = True) -> None:
        """
        Enable or disable caching completely.

        Because caches are repository-specific, disabling the cache
        cannot immediately clear all cached objects, but each cache will
        be cleared on the next attempt to update anything in it.
        """
        return option(Option.ENABLE_CACHING, value)

    def disable_pack_keep_file_checks(self, value: bool = True) -> None:
        """
        This will cause .keep file existence checks to be skipped when
        accessing packfiles, which can help performance with remote
        filesystems.
        """
        return option(Option.DISABLE_PACK_KEEP_FILE_CHECKS, value)

    def cache_max_size(self, value: int) -> None:
        """
        Set the maximum total data size that will be cached in memory
        across all repositories before libgit2 starts evicting objects
        from the cache.  This is a soft limit, in that the library might
        briefly exceed it, but will start aggressively evicting objects
        from cache when that happens.  The default cache size is 256MB.
        """
        return option(Option.SET_CACHE_MAX_SIZE, value)

    def cache_object_limit(
        self, object_type: pygit2.enums.ObjectType, value: int
    ) -> None:
        """
        Set the maximum data size for the given type of object to be
        considered eligible for caching in memory.  Setting to value to
        zero means that that type of object will not be cached.
        Defaults to 0 for enums.ObjectType.BLOB (i.e. won't cache blobs)
        and 4k for COMMIT, TREE, and TAG.
        """
        return option(Option.SET_CACHE_OBJECT_LIMIT, object_type, value)

    @property
    def ssl_cert_file(self) -> str | bytes | None:
        """TLS certificate file path."""
        return self._ssl_cert_file

    @ssl_cert_file.setter
    def ssl_cert_file(self, value: str | bytes) -> None:
        """Set the TLS cert file path."""
        self.set_ssl_cert_locations(value, self._ssl_cert_dir)

    @ssl_cert_file.deleter
    def ssl_cert_file(self) -> None:
        """Reset the TLS cert file path."""
        self.ssl_cert_file = self._default_tls_verify_paths.cafile  # type: ignore[union-attr]

    @property
    def ssl_cert_dir(self) -> str | bytes | None:
        """TLS certificates lookup directory path."""
        return self._ssl_cert_dir

    @ssl_cert_dir.setter
    def ssl_cert_dir(self, value: str | bytes) -> None:
        """Set the TLS certificate lookup folder."""
        self.set_ssl_cert_locations(self._ssl_cert_file, value)

    @ssl_cert_dir.deleter
    def ssl_cert_dir(self) -> None:
        """Reset the TLS certificate lookup folder."""
        self.ssl_cert_dir = self._default_tls_verify_paths.capath  # type: ignore[union-attr]

    @overload
    def set_ssl_cert_locations(
        self, cert_file: str | bytes | None, cert_dir: str | bytes
    ) -> None: ...
    @overload
    def set_ssl_cert_locations(
        self, cert_file: str | bytes, cert_dir: str | bytes | None
    ) -> None: ...
    def set_ssl_cert_locations(
        self, cert_file: str | bytes | None, cert_dir: str | bytes | None
    ) -> None:
        """
        Set the SSL certificate-authority locations.

        - `cert_file` is the location of a file containing several
          certificates concatenated together.
        - `cert_dir` is the location of a directory holding several
          certificates, one per file.

        Either parameter may be `NULL`, but not both.
        """
        option(Option.SET_SSL_CERT_LOCATIONS, cert_file, cert_dir)
        self._ssl_cert_file = cert_file
        self._ssl_cert_dir = cert_dir

    @property
    def template_path(self) -> str | None:
        """Get or set the default template path for new repositories"""
        return option(Option.GET_TEMPLATE_PATH)

    @template_path.setter
    def template_path(self, value: str | bytes) -> None:
        option(Option.SET_TEMPLATE_PATH, value)

    @property
    def user_agent(self) -> str | None:
        """Get or set the user agent string for network operations"""
        return option(Option.GET_USER_AGENT)

    @user_agent.setter
    def user_agent(self, value: str | bytes) -> None:
        option(Option.SET_USER_AGENT, value)

    @property
    def user_agent_product(self) -> str | None:
        """Get or set the user agent product name"""
        return option(Option.GET_USER_AGENT_PRODUCT)

    @user_agent_product.setter
    def user_agent_product(self, value: str | bytes) -> None:
        option(Option.SET_USER_AGENT_PRODUCT, value)

    def set_ssl_ciphers(self, ciphers: str | bytes) -> None:
        """Set the SSL ciphers to use for HTTPS connections"""
        option(Option.SET_SSL_CIPHERS, ciphers)

    def enable_strict_object_creation(self, value: bool = True) -> None:
        """Enable or disable strict object creation validation"""
        option(Option.ENABLE_STRICT_OBJECT_CREATION, value)

    def enable_strict_symbolic_ref_creation(self, value: bool = True) -> None:
        """Enable or disable strict symbolic reference creation validation"""
        option(Option.ENABLE_STRICT_SYMBOLIC_REF_CREATION, value)

    def enable_ofs_delta(self, value: bool = True) -> None:
        """Enable or disable offset delta encoding"""
        option(Option.ENABLE_OFS_DELTA, value)

    def enable_fsync_gitdir(self, value: bool = True) -> None:
        """Enable or disable fsync for git directory operations"""
        option(Option.ENABLE_FSYNC_GITDIR, value)

    def enable_strict_hash_verification(self, value: bool = True) -> None:
        """Enable or disable strict hash verification"""
        option(Option.ENABLE_STRICT_HASH_VERIFICATION, value)

    def enable_unsaved_index_safety(self, value: bool = True) -> None:
        """Enable or disable unsaved index safety checks"""
        option(Option.ENABLE_UNSAVED_INDEX_SAFETY, value)

    def enable_http_expect_continue(self, value: bool = True) -> None:
        """Enable or disable HTTP Expect/Continue for large pushes"""
        option(Option.ENABLE_HTTP_EXPECT_CONTINUE, value)

    @property
    def windows_sharemode(self) -> int:
        """Get or set the Windows share mode for opening files"""
        return option(Option.GET_WINDOWS_SHAREMODE)

    @windows_sharemode.setter
    def windows_sharemode(self, value: int) -> None:
        option(Option.SET_WINDOWS_SHAREMODE, value)

    @property
    def pack_max_objects(self) -> int:
        """Get or set the maximum number of objects in a pack"""
        return option(Option.GET_PACK_MAX_OBJECTS)

    @pack_max_objects.setter
    def pack_max_objects(self, value: int) -> None:
        option(Option.SET_PACK_MAX_OBJECTS, value)

    @property
    def owner_validation(self) -> bool:
        """Get or set repository directory ownership validation"""
        return option(Option.GET_OWNER_VALIDATION)

    @owner_validation.setter
    def owner_validation(self, value: bool) -> None:
        option(Option.SET_OWNER_VALIDATION, value)

    def set_odb_packed_priority(self, priority: int) -> None:
        """Set the priority for packed ODB backend (default 1)"""
        option(Option.SET_ODB_PACKED_PRIORITY, priority)

    def set_odb_loose_priority(self, priority: int) -> None:
        """Set the priority for loose ODB backend (default 2)"""
        option(Option.SET_ODB_LOOSE_PRIORITY, priority)

    @property
    def extensions(self) -> list[str]:
        """Get the list of enabled extensions"""
        return option(Option.GET_EXTENSIONS)

    def set_extensions(self, extensions: list[str]) -> None:
        """Set the list of enabled extensions"""
        option(Option.SET_EXTENSIONS, extensions, len(extensions))

    @property
    def homedir(self) -> str | None:
        """Get or set the home directory"""
        return option(Option.GET_HOMEDIR)

    @homedir.setter
    def homedir(self, value: str | bytes) -> None:
        option(Option.SET_HOMEDIR, value)

    @property
    def server_connect_timeout(self) -> int:
        """Get or set the server connection timeout in milliseconds"""
        return option(Option.GET_SERVER_CONNECT_TIMEOUT)

    @server_connect_timeout.setter
    def server_connect_timeout(self, value: int) -> None:
        option(Option.SET_SERVER_CONNECT_TIMEOUT, value)

    @property
    def server_timeout(self) -> int:
        """Get or set the server timeout in milliseconds"""
        return option(Option.GET_SERVER_TIMEOUT)

    @server_timeout.setter
    def server_timeout(self, value: int) -> None:
        option(Option.SET_SERVER_TIMEOUT, value)
